"""This is ported py-version of the pru-code.

Goals:

- stay close to original code-base
- offer a comparison for the tests
- step 1 to a virtualization of emulation.

NOTE: DO NOT OPTIMIZE -> stay close to original code-base

"""

from shepherd_core.data_models.base.calibration import CalibrationEmulator
from shepherd_core.data_models.content.energy_environment import EnergyDType
from shepherd_core.data_models.content.virtual_harvester_config import HarvesterPRUConfig
from shepherd_core.data_models.content.virtual_source_config import ConverterPRUConfig
from shepherd_core.data_models.content.virtual_source_config import VirtualSourceConfig
from shepherd_core.data_models.content.virtual_storage_config import StoragePRUConfig

from .virtual_converter_model import PruCalibration
from .virtual_converter_model import VirtualConverterModel
from .virtual_harvester_model import VirtualHarvesterModel


class VirtualSourceModel:
    """part of sampling.c."""

    def __init__(
        self,
        vsrc: VirtualSourceConfig | None,
        cal_emu: CalibrationEmulator,
        dtype_in: EnergyDType = EnergyDType.ivsample,
        window_size: int | None = None,
        voltage_step_V: float | None = None,
        *,
        log_intermediate: bool = False,
    ) -> None:
        self._cal_emu: CalibrationEmulator = cal_emu
        self._cal_pru: PruCalibration = PruCalibration(cal_emu)
        self.cfg_src = VirtualSourceConfig() if vsrc is None else vsrc

        # init Harvester
        cfg_hrv = HarvesterPRUConfig.from_vhrv(
            self.cfg_src.harvester,
            for_emu=True,
            dtype_in=dtype_in,
            window_size=window_size,
            voltage_step_V=voltage_step_V,
        )
        self.hrv: VirtualHarvesterModel = VirtualHarvesterModel(cfg_hrv)

        # init Converters
        cfg_cnv = ConverterPRUConfig.from_vsrc(
            self.cfg_src,
            dtype_in=dtype_in,
            log_intermediate_node=log_intermediate,
        )
        cfg_storage = StoragePRUConfig.from_vstorage(self.cfg_src.storage, optimize_clamp=True)
        self.cnv: VirtualConverterModel = VirtualConverterModel(cfg_cnv, self._cal_pru, cfg_storage)

        # state for simulation
        self.W_inp_fWs: float = 0.0
        self.W_out_fWs: float = 0.0

    def iterate_sampling(self, V_inp_uV: int = 0, I_inp_nA: int = 0, I_out_nA: int = 0) -> int:
        """TEST-SIMPLIFICATION - code below is not part of pru-code.

        It originates from sample_emulator() in sampling.c.

        :param V_inp_uV:
        :param I_inp_nA:
        :param I_out_nA:
        :return:
        """
        V_inp_uV, I_inp_nA = self.hrv.ivcurve_sample(V_inp_uV, I_inp_nA)

        P_inp_fW = self.cnv.calc_inp_power(V_inp_uV, I_inp_nA)

        # fake ADC read
        A_out_raw = self._cal_emu.adc_C_A.si_to_raw(I_out_nA * 10**-9)
        P_out_fW = self.cnv.calc_out_power(A_out_raw)

        V_mid_uV = self.cnv.update_cap_storage()
        V_out_raw = self.cnv.update_states_and_output()
        V_out_uV = int(self._cal_emu.dac_V_A.raw_to_si(V_out_raw) * 10**6)

        self.W_inp_fWs += P_inp_fW
        self.W_out_fWs += P_out_fW

        # feedback path - important for boost-less circuits
        if self.cnv.feedback_to_hrv:
            self.hrv.voltage_set_uV = self.cnv.V_input_request_uV

        return V_mid_uV if self.cnv.get_state_log_intermediate() else V_out_uV



"""Simulate behavior of virtual harvester algorithms.

The simulation recreates an observer-cape and the virtual harvester
- input = hdf5-file with an ivcurve
- output = optional as hdf5-file

The output file can be analyzed and plotted with shepherds tool suite.
"""

from contextlib import ExitStack
from pathlib import Path

from tqdm import tqdm

from shepherd_core.data_models.base.calibration import CalibrationHarvester
from shepherd_core.data_models.content.virtual_harvester_config import HarvesterPRUConfig
from shepherd_core.data_models.content.virtual_harvester_config import VirtualHarvesterConfig
from shepherd_core.reader import Reader
from shepherd_core.writer import Writer

from .virtual_harvester_model import VirtualHarvesterModel


def simulate_harvester(
    config: VirtualHarvesterConfig, path_input: Path, path_output: Path | None = None
) -> float:
    """Simulate behavior of virtual harvester algorithms.

    Fn return the harvested energy.
    """
    with ExitStack() as stack:
        file_inp = Reader(path_input, verbose=False)
        stack.enter_context(file_inp)
        cal_inp = file_inp.get_calibration_data()

        if path_output:
            cal_hrv = CalibrationHarvester()
            file_out = Writer(
                path_output, cal_data=cal_hrv, mode="harvester", verbose=False, force_overwrite=True
            )
            stack.enter_context(file_out)
            cal_out = file_out.get_calibration_data()
            file_out.store_hostname("hrv_sim_" + config.name)

        hrv_pru = HarvesterPRUConfig.from_vhrv(
            config,
            for_emu=True,
            dtype_in=file_inp.get_datatype(),
            window_size=file_inp.get_window_samples(),
            voltage_step_V=file_inp.get_voltage_step(),
        )
        hrv = VirtualHarvesterModel(hrv_pru)
        e_out_Ws = 0.0

        for _t, v_inp, i_inp in tqdm(
            file_inp.read(is_raw=True), total=file_inp.chunks_n, desc="Chunk", leave=False
        ):
            v_uV = cal_inp.voltage.raw_to_si(v_inp) * 1e6
            i_nA = cal_inp.current.raw_to_si(i_inp) * 1e9
            length = min(v_uV.size, i_nA.size)
            for _n in range(length):
                v_uV[_n], i_nA[_n] = hrv.ivcurve_sample(
                    _voltage_uV=int(v_uV[_n]), _current_nA=int(i_nA[_n])
                )
            e_out_Ws += (v_uV * i_nA).sum() * 1e-15 * file_inp.sample_interval_s
            if path_output:
                v_out = cal_out.voltage.si_to_raw(v_uV / 1e6)
                i_out = cal_out.current.si_to_raw(i_nA / 1e9)
                file_out.append_iv_data_raw(_t, v_out, i_out)

    return e_out_Ws



"""Config for testbed experiments."""

from collections.abc import Iterable
from datetime import datetime
from datetime import timedelta
from typing import Annotated

from pydantic import Field
from pydantic import model_validator
from typing_extensions import Self

from shepherd_core.config import config
from shepherd_core.data_models.base.content import NameStr
from shepherd_core.data_models.base.content import SafeStr
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.base.timezone import local_now
from shepherd_core.data_models.testbed.target import Target
from shepherd_core.data_models.testbed.testbed import Testbed
from shepherd_core.version import version

from .observer_features import SystemLogging
from .target_config import TargetConfig

# defaults (pre-init complex types)
sys_log_all = SystemLogging()  # = all active


class Experiment(ShpModel, title="Config of an Experiment"):
    """Config for experiments on the testbed emulating energy environments for target nodes."""

    # General Properties
    name: NameStr
    description: Annotated[SafeStr | None, Field(description="Required for public instances")] = (
        None
    )
    comment: SafeStr | None = None

    # feedback
    email_results: bool = True

    sys_logging: SystemLogging = sys_log_all

    # schedule
    time_start: datetime | None = None  # = ASAP
    duration: timedelta | None = None  # = till EOF

    # targets
    target_configs: Annotated[list[TargetConfig], Field(min_length=1, max_length=128)]

    # debug
    lib_ver: str | None = version

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        self._validate_observers(self.target_configs)
        self._validate_targets(self.target_configs)
        if self.duration and self.duration.total_seconds() < 0:
            raise ValueError("Duration of experiment can't be negative.")
        return self

    @staticmethod
    def _validate_targets(configs: Iterable[TargetConfig]) -> None:
        target_ids: list[int] = []
        custom_ids: list[int] = []
        for _config in configs:
            for _id in _config.target_IDs:
                target_ids.append(_id)
                if config.VALIDATE_INFRA:
                    Target(id=_id)
                    # ⤷ this can raise exception for non-existing targets
            if _config.custom_IDs is not None:
                custom_ids = custom_ids + _config.custom_IDs[: len(_config.target_IDs)]
            else:
                custom_ids = custom_ids + _config.target_IDs
        if len(target_ids) > len(set(target_ids)):
            raise ValueError("Target-ID used more than once in Experiment!")
        if len(target_ids) > len(set(custom_ids)):
            raise ValueError("Custom Target-ID are faulty (some form of id-collisions)!")

    @staticmethod
    def _validate_observers(configs: Iterable[TargetConfig]) -> None:
        if not config.VALIDATE_INFRA:
            return
        testbed = Testbed()
        target_ids = [_id for _config in configs for _id in _config.target_IDs]
        obs_ids = [testbed.get_observer(_id).id for _id in target_ids]
        if len(target_ids) > len(set(obs_ids)):
            raise ValueError(
                "Observer is used more than once in Experiment -> only 1 target per observer!"
            )

    def get_target_ids(self) -> list:
        return [_id for _config in self.target_configs for _id in _config.target_IDs]

    def get_target_config(self, target_id: int) -> TargetConfig:
        for _config in self.target_configs:
            if target_id in _config.target_IDs:
                return _config
        # gets already caught in target_config - but keep:
        msg = f"Target-ID {target_id} was not found in Experiment '{self.name}'"
        raise ValueError(msg)

    def folder_name(self, custom_date: datetime | None = None) -> str:
        # TODO: custom date should not overrule time_start
        date = custom_date if custom_date is not None else self.time_start
        timestamp = local_now() if date is None else date
        timestrng = timestamp.strftime("%Y-%m-%d_%H-%M-%S")
        # ⤷ closest to ISO 8601, avoids ":"
        return f"{timestrng}_{self.name.replace(' ', '_')}"



"""Demonstrate behavior of Virtual Source Algorithms.

- simulation combines a range of
    - virtual sources
    - simulation durations
    - harvesting voltages
- harvesting has a duty cycle (steps visible in plot)
- MCU switches between active/sleep depending on power-good-signal
- results are plotted

"""

from itertools import product

import matplotlib.pyplot as plt
import numpy as np
from shepherd_core.data_models import VirtualSourceConfig
from shepherd_core.vsource import VirtualSourceModel

from shepherd_core import CalibrationEmulator
from shepherd_core import WebClient

# config simulation
sample_dur_list = [50_000, 500_000]
v_hrv_mV_list = [1200, 3300]
charge_rate = 0.5

src_list = [
    "direct",
    "diode+capacitor",
    "diode+resistor+capacitor",
    "BQ25504",
    "BQ25504s",
    "BQ25570",
    "BQ25570s",
]

I_mcu_sleep_A = 200e-9
I_mcu_active_A = 1e-3

# For online-queries the lib can be connected to the testbed-server.
# NOTE: there are 3 states:
# - unconnected -> demo-fixtures are queried (locally)
# - connected -> publicly available data is queried online
# - logged in with valid token -> also private data is queried online
do_connect = False
if do_connect:
    WebClient()

for vs_name, v_hrv_mV, samples in product(src_list, v_hrv_mV_list, sample_dur_list):
    # prepare simulation
    ts = np.arange(0, samples * 10e-6, 10e-6)
    vcaps = np.empty((samples, 3))

    src_config = VirtualSourceConfig(name=vs_name)
    cal_emu = CalibrationEmulator()
    src = VirtualSourceModel(src_config, cal_emu, log_intermediate=True)

    I_out_nA = int(I_mcu_sleep_A * 10**9)
    N_good = 0
    for i in range(samples):
        # Harvest at 100uA
        if (i % 1_000) < charge_rate * 1_000:
            V_out_uV = src.iterate_sampling(
                V_inp_uV=v_hrv_mV * 10**3,
                I_inp_nA=100_000,
                I_out_nA=I_out_nA,
            )
        else:
            V_out_uV = src.iterate_sampling(I_out_nA=I_out_nA)

        # listen to power-good signal
        if src.cnv.get_power_good():
            I_out_nA = int(I_mcu_active_A * 10**9)
            N_good += 1
        else:
            I_out_nA = int(I_mcu_sleep_A * 10**9)

        # store sim-results
        vcaps[i] = [
            src.cnv.get_V_intermediate_uV() * 1e-6,
            V_out_uV * 1e-6,
            src.cnv.get_power_good(),
        ]

    # visualize results
    pwr_good_ratio = round(100 * N_good / samples, 3)
    t_s = samples // 100
    print(f"Power-Good-Ratio = {pwr_good_ratio} % (for {vs_name}, {v_hrv_mV} mV, {t_s} ms)")

    fig = plt.figure(figsize=(20, 8))
    plt.plot(ts, vcaps)
    plt.suptitle(
        f"VSrc-Simulation [{vs_name}, V_inp={v_hrv_mV}mV, "
        f"duty={charge_rate}, duration={t_s}ms] "
        f"-> pwr-good-ratio={pwr_good_ratio} %"
    )
    plt.xlabel("Runtime [s]")
    plt.ylabel("Voltages [V]")
    plt.legend(["V_Cap", "V_Out", "PGood"], loc="upper right")
    plt.tight_layout()
    plt.savefig(f"./vsource_sim_{vs_name}_{v_hrv_mV}mV_{t_s}ms.png")
    plt.close(fig)



"""Script for generating YAML-fixture for virtual storage."""

import sys
from pathlib import Path

import yaml
from virtual_storage_config import VirtualStorageConfig

from shepherd_core import local_now
from shepherd_core.data_models import Wrapper
from shepherd_core.logger import log

dsc_ideal = "Model of an ideal Capacitor (true to spec, no losses)"
dsc_tantal = "Tantal-Capacitor similar to ideal Model, but with R_leak & R_series"
dsc_mlcc = "MLCC-Capacitor with R_leak & R_series and planned DC-Bias-Effect"
dsc_super = "SuperCapacitor with typically 1000 hours / 500 k cycles (not modeled)"

# Ideal Capacitor, E6 row 10 to 1000 uF
# typical voltage-ratings: 2.5, 4.0, 6.3, 10, 16, 20 V
E6: list[int] = [10, 15, 22, 33, 47, 68, 100, 150, 220, 330, 470, 680, 1000]
fixture_ideal: list[VirtualStorageConfig] = [
    VirtualStorageConfig.capacitor(C_uF=_v, V_rated=10.0, description=dsc_ideal) for _v in E6
]

# Tantal Capacitors, E6 row
# ⤷ verified with AVX TAJB107M006RNJ
#       - Tantal, 100 uF, 6V3, 20%, 1411 package -> 100 uF measured
#       - R_series taken from Datasheet
#       - R_leak = 1 MOhm in datasheet (6.3V/6.3uA), but ~2 MOhm in experiment
# see https://github.com/orgua/bq_characteristics/tree/main/eval_kit_behavior_var1#capacitor
fixture_tantal: list[VirtualStorageConfig] = [
    VirtualStorageConfig.capacitor(
        C_uF=10,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 10,
        R_series_Ohm=3.0,
        name="AVX TAJB106x006",  # Note: small x is * in datasheet
        description=dsc_tantal,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=15,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 15,
        R_series_Ohm=2.0,
        name="AVX TAJB156x006",
        description=dsc_tantal,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=22,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 22,
        R_series_Ohm=2.5,
        name="AVX TAJB226x006",
        description=dsc_tantal,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=33,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 33,
        R_series_Ohm=2.2,
        name="AVX TAJB336x006",
        description=dsc_tantal,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=47,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 47,
        R_series_Ohm=2,
        name="AVX TAJB476x006",
        description=dsc_tantal,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=68,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 68,
        R_series_Ohm=0.9,
        name="AVX TAJB686x006",
        description=dsc_tantal,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=100,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 100,
        R_series_Ohm=1.7,
        name="AVX TAJB107x006",
        description=dsc_tantal,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=150,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 150,
        R_series_Ohm=1.3,
        name="AVX TAJC157x006",
        description=dsc_tantal,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=220,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 220,
        R_series_Ohm=1.2,
        name="AVX TAJC227x006",
        description=dsc_tantal,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=330,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 330,
        R_series_Ohm=0.5,
        name="AVX TAJC337x006",
        description=dsc_tantal,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=470,
        V_rated=6.3,
        R_leak_Ohm=196e6 / 470,
        R_series_Ohm=0.4,
        name="AVX TAJD477x006",
        description=dsc_tantal,
    ),
]

# MLCC
# ⤷ verified with Taiyo Yuden JMK316ABJ107ML-T
#       - MLCC, 100uF, 6V3, 20%, X5R, 1206 package -> 74 uF measured
#       - Insulation Resistance (min) 100 MΩ·μF (datasheet), 97.8 MΩ measured
# https://github.com/orgua/bq_characteristics/tree/main/eval_kit_behavior_var1/data_capacitor
# BQ25570EVM uses Murata GRM43SR60J107ME20L
#       - MLCC, 100uF, 6V3, 20%, X5R, 1812 package -> 79 uF measured
#       - murata-DB does not know this part, but direct substitute is GRM31CR60J107MEA8
# https://pim.murata.com/en-global/pim/details/?partNum=GRM31CR60J107MEA8%23
# TODO: add DC-Bias
fixture_mlcc: list[VirtualStorageConfig] = [
    VirtualStorageConfig.capacitor(
        C_uF=10, V_rated=6.3, R_leak_Ohm=97.8e6 / 10, description=dsc_mlcc
    ),
    VirtualStorageConfig.capacitor(
        C_uF=33, V_rated=6.3, R_leak_Ohm=97.8e6 / 33, description=dsc_mlcc
    ),
    VirtualStorageConfig.capacitor(
        C_uF=47, V_rated=6.3, R_leak_Ohm=97.8e6 / 47, description=dsc_mlcc
    ),
    VirtualStorageConfig.capacitor(
        C_uF=100, V_rated=6.3, R_leak_Ohm=97.8e6 / 100, description=dsc_mlcc
    ),
]

# SuperCap
fixture_super: list[VirtualStorageConfig] = [
    VirtualStorageConfig.capacitor(
        C_uF=25e6,
        V_rated=3.0,
        R_leak_Ohm=3 / 55e-6,
        R_series_Ohm=17e-3,
        name="Maxwell BCAP0025 P300 X11",
        description=dsc_super,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=12e6,
        V_rated=6.0,
        R_leak_Ohm=6 / 80e-6,
        R_series_Ohm=90e-3,
        name="Abracon ADCM-S06R0SA126RB",
        description=dsc_super,
    ),
    VirtualStorageConfig.capacitor(
        C_uF=7.5e6,
        V_rated=5.5,
        R_leak_Ohm=6 / 78e-6,
        R_series_Ohm=90e-3,
        name="AVX SCMT32F755SRBA0 ",
        description=dsc_super,
    ),
]

fixture_lipo: list[VirtualStorageConfig] = [
    # LiPo-Coin-cells 5.4mm, https://www.lipobatteries.net/3-8v-rechargeable-mini-button-lipo-coin-cell-battery/
    VirtualStorageConfig.lipo(
        q_mAh=80, name="LPM1254", description="LiPo-Coin-Cell 80mAh, 1 cell, w=12mm, d=5.4mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=65, name="LPM1154", description="LiPo-Coin-Cell 65mAh, 1 cell, w=11mm, d=5.4mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=50, name="LPM1054", description="LiPo-Coin-Cell 50mAh, 1 cell, w=10mm, d=5.4mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=40, name="LPM0954", description="LiPo-Coin-Cell 40mAh, 1 cell, w=9mm, d=5.4mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=35, name="LPM0854", description="LiPo-Coin-Cell 35mAh, 1 cell, w=8mm, d=5.4mm"
    ),
    # LiPo-Coin-cells 4.0mm, https://www.lipobatteries.net/3-8v-rechargeable-mini-button-lipo-coin-cell-battery/
    VirtualStorageConfig.lipo(
        q_mAh=55, name="LPM1240", description="LiPo-Coin-Cell 55mAh, 1 cell, w=12mm, d=4.0mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=45, name="LPM1140", description="LiPo-Coin-Cell 45mAh, 1 cell, w=11mm, d=4.0mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=35, name="LPM1040", description="LiPo-Coin-Cell 35mAh, 1 cell, w=10mm, d=4.0mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=30, name="LPM0940", description="LiPo-Coin-Cell 30mAh, 1 cell, w=9mm, d=4.0mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=18, name="LPM0840", description="LiPo-Coin-Cell 18mAh, 1 cell, w=8mm, d=4.0mm"
    ),
    # small LiPos, https://www.lipobatteries.net/lipo-batteries-within-100mah/
    VirtualStorageConfig.lipo(
        q_mAh=12, name="LP151020", description="LiPo-Pouch 12mAh, 1 cell, 20x10x1.5mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=15, name="LP251212", description="LiPo-Pouch 15mAh, 1 cell, 12x12x2.5mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=22, name="LP500522", description="LiPo-Pouch 20mAh, 1 cell, 22x05x5.0mm"
    ),
    VirtualStorageConfig.lipo(
        q_mAh=22, name="LP271015", description="LiPo-Pouch 22mAh, 1 cell, 15x10x2.7mm"
    ),
    # Example from the paper
    VirtualStorageConfig.lipo(q_mAh=860, name="PL-383562"),
]

fixture_lead: list[VirtualStorageConfig] = [
    # Example from the paper
    VirtualStorageConfig.lead_acid(q_mAh=1200, name="LEOCH_LP12-1.2AH"),
]


if __name__ == "__main__":
    path_here = Path(__file__).parent.absolute()
    path_db = path_here  # .parent / "shepherd_core/shepherd_core/data_models/content"

    if not path_db.exists() or not path_db.is_dir():
        log.error("Path to db must exist and be a directory!")
        sys.exit(1)

    fixtures: dict[str, list[VirtualStorageConfig]] = {
        "ideal": fixture_ideal,
        "tantal": fixture_tantal,
        "mlcc": fixture_mlcc,
        "super": fixture_super,
        "lipo": fixture_lipo,
        "lead": fixture_lead,
    }

    for name, fixture in fixtures.items():
        file_path = path_db / f"virtual_storage_fixture_{name}.yaml"
        if file_path.exists():
            log.warning("File %s already exists! -> will skip", file_path.name)
        models_wrap = []
        for model in fixture:
            model_dict = (
                model.model_dump()
            )  # exclude_unset=True, exclude_defaults=True, include={"id",})
            model_wrap = Wrapper(
                datatype=type(model).__name__,
                created=local_now(),
                comment=f"created by script '{Path(__file__).name}'",
                parameters=model_dict,
            )
            models_wrap.append(model_wrap.model_dump(exclude_unset=True, exclude_defaults=True))

        models_yaml = yaml.safe_dump(models_wrap, default_flow_style=False, sort_keys=False)
        with file_path.open("w") as f:
            f.write(models_yaml)



"""Demonstrate behavior of Virtual Harvester Algorithms.

- simulation is based on IVTrace derived from a solar-isc-voc-recording during a jogging-trip
- harvesting is done by various algorithms and preconfigured virtual harvesters
- results are printed on console (harvested energy)

Output:
E_out = 0.000 mWs -> cv20
E_out = 17.143 mWs -> cv10
E_out = 17.384 mWs -> mppt_voc
E_out = 17.249 mWs -> mppt_bq_solar
E_out = 13.954 mWs -> mppt_bq_thermoelectric
E_out = 15.001 mWs -> mppt_po
E_out = 17.811 mWs -> mppt_opt

"""

from pathlib import Path

from shepherd_core.data_models import VirtualHarvesterConfig
from shepherd_core.vsource import simulate_harvester

from shepherd_core import Reader
from shepherd_data import ivonne

# config simulation
sim_duration = 32
file_ivonne = Path(__file__).parents[3] / "shepherd_data/examples/jogging_10m.iv"
file_ivcurve = Path(__file__).parent / "jogging_ivcurve.h5"

if not file_ivonne.exists():
    raise FileNotFoundError("Input-File not found - check path")

hrv_list = [
    "cv20",
    # ⤷ fails due to lower solar voltage
    "cv10",
    "mppt_voc",
    "mppt_bq_solar",
    # ⤷ bq needs 16 s to start -> bad performance for this demo
    "mppt_bq_thermoelectric",
    # ⤷ thermoelectric setpoint -> bad performance for solar
    "mppt_po",
    "mppt_opt",
]

save_files: bool = True

# convert IVonne to IVCurve
if not file_ivcurve.exists():
    with ivonne.Reader(file_ivonne) as db:
        db.convert_2_ivsurface(file_ivcurve, duration_s=sim_duration)

# Input Statistics
with Reader(file_ivcurve, verbose=False) as file:
    window_size = file.get_window_samples()
    I_in_max = 0.0
    for _t, _v, _i in file.read():
        I_in_max = max(I_in_max, _i.max())
    print(
        f"Input-file: \n"
        f"\tE_in = {file.energy() * 1e3:.3f} mWs (not representative)\n"
        f"\tI_in_max = {I_in_max * 1e3:.3f} mA\n"
        f"\twindow_size = {window_size} n\n",
    )

# Simulation
for hrv_name in hrv_list:
    file_output = file_ivcurve.with_stem(file_ivcurve.stem + "_" + hrv_name) if save_files else None
    E_out_Ws = simulate_harvester(
        config=VirtualHarvesterConfig(name=hrv_name),
        path_input=file_ivcurve,
        path_output=file_output,
    )
    print(f"E_out = {E_out_Ws * 1e3:.3f} mWs -> {hrv_name}")



import os
from pathlib import Path

import numpy as np
import pytest
from pydantic import ValidationError
from shepherd_core.data_models.base.cal_measurement import CalMeasurementCape
from shepherd_core.data_models.base.calibration import CalibrationCape
from shepherd_core.data_models.base.calibration import CalibrationEmulator
from shepherd_core.data_models.base.calibration import CalibrationHarvester
from shepherd_core.data_models.base.calibration import CalibrationPair
from shepherd_core.data_models.base.calibration import CalibrationSeries
from shepherd_core.data_models.base.calibration import CapeData
from shepherd_core.data_models.base.content import ContentModel


def test_base_model_cape_data() -> None:
    CapeData(serial_number="xyz1")


def test_base_model_cape_data_fail() -> None:
    with pytest.raises(ValidationError):
        CapeData()


def test_base_model_cal_pair_conv() -> None:
    cal = CalibrationPair(gain=4.9)
    val_raw = 500
    val_si = cal.raw_to_si(val_raw)
    val_rbw = cal.si_to_raw(val_si)
    assert val_raw == val_rbw


def test_base_model_cal_pair_conv2() -> None:
    cal = CalibrationPair(gain=44)
    rng = np.random.default_rng()
    val_raw = rng.integers(low=0, high=2000, size=20)
    val_si = cal.raw_to_si(val_raw)
    val_rbw = cal.si_to_raw(val_si)
    assert val_raw.size == val_rbw.size


def test_base_model_cal_series_min() -> None:
    CalibrationSeries()


def test_base_model_cal_series_all() -> None:
    cal = CalibrationCape()
    _ = CalibrationSeries.from_cal(cal.harvester)
    _ = CalibrationSeries.from_cal(cal.emulator)
    _ = CalibrationSeries.from_cal(cal.emulator, emu_port_a=True)
    _ = CalibrationSeries.from_cal(cal.emulator, emu_port_a=False)


def test_base_model_cal_hrv_min() -> None:
    cal = CalibrationHarvester()
    cs = cal.export_for_sysfs()
    assert len(cs) == 6


def test_base_model_cal_emu_min() -> None:
    cal = CalibrationEmulator()
    cs = cal.export_for_sysfs()
    assert len(cs) == 6


def test_base_model_cal_cape_bytestr() -> None:
    cal1 = CalibrationCape()
    cb = cal1.to_bytestr()
    cal2 = CalibrationCape.from_bytestr(cb)
    assert cal1.get_hash() == cal2.get_hash()


def test_base_model_cal_cape_bytestr_with_cape_data() -> None:
    cal1 = CalibrationCape(cape=CapeData(serial_number="123"))
    cb = cal1.to_bytestr()
    cal2 = CalibrationCape.from_bytestr(cb, cal1.cape)
    assert cal1.get_hash() == cal2.get_hash()


def test_base_model_cal_cape_example(tmp_path: Path) -> None:
    cal0 = CalMeasurementCape()
    path1 = Path(__file__).resolve().with_name("example_cal_data.yaml")
    cal1 = CalibrationCape.from_file(path1)
    path2 = tmp_path / "cal_data_new.yaml"
    cal1.to_file(path2)
    cal2 = CalibrationCape.from_file(path2)
    assert cal0.get_hash() != cal1.get_hash()
    assert cal1.get_hash() == cal2.get_hash()


def test_base_model_cal_hrv_fault() -> None:
    path = Path(__file__).resolve().with_name("example_cal_data_faulty.yaml")
    cal = CalibrationCape.from_file(path)
    with pytest.raises(ValueError):  # noqa: PT011
        _ = cal.harvester.export_for_sysfs()


def test_base_model_cal_emu_fault() -> None:
    path = Path(__file__).resolve().with_name("example_cal_data_faulty.yaml")
    cal = CalibrationCape.from_file(path)
    with pytest.raises(ValueError):  # noqa: PT011
        _ = cal.emulator.export_for_sysfs()


def test_base_model_cal_meas_min() -> None:
    cm1 = CalMeasurementCape()
    cal1 = cm1.to_cal()
    cal2 = CalibrationCape()
    assert cal1.get_hash() == cal2.get_hash()


def test_base_model_cal_meas_example() -> None:
    path1 = Path(__file__).resolve().with_name("example_cal_meas.yaml")
    cm1 = CalMeasurementCape.from_file(path1)
    cal1 = cm1.to_cal()
    cal2 = CalibrationCape()
    assert cal1.get_hash() != cal2.get_hash()


def test_base_model_cal_meas_fault_correlation() -> None:
    path = Path(__file__).resolve().with_name("example_cal_meas_faulty1.yaml")
    with pytest.raises(ValidationError):
        _ = CalMeasurementCape.from_file(path)


def test_base_model_cal_meas_fault_no_pairs() -> None:
    path = Path(__file__).resolve().with_name("example_cal_meas_faulty2.yaml")
    with pytest.raises(ValidationError):
        _ = CalMeasurementCape.from_file(path)


def test_base_model_content_str_repr() -> None:
    content = ContentModel(name="tricky", owner="peter", group="work")
    print(content)
    assert str(content) == "tricky"


def test_base_model_content_public_group() -> None:
    _ = ContentModel(
        name="tricky",
        description="fake",
        owner="peter",
        group="work",
        visible2group=True,
    )
    with pytest.raises(ValidationError):
        _ = ContentModel(name="tricky", owner="peter", group="work", visible2group=True)


def test_base_model_content_public_all() -> None:
    _ = ContentModel(
        name="tricky", description="fake", owner="peter", group="work", visible2all=True
    )
    with pytest.raises(ValidationError):
        _ = ContentModel(name="tricky", owner="peter", group="work", visible2all=True)


# ContentModel below is used as inheritor of ShpModel


def test_base_model_shepherd_scheme(tmp_path: Path) -> None:
    os.chdir(tmp_path)
    ContentModel.schema_to_file(tmp_path / "schema.yaml")


def test_base_model_shepherd_min(tmp_path: Path) -> None:
    content = ContentModel(name="tricky", owner="peter", group="work")
    content.to_file(tmp_path / "content1.yaml", minimal=True)
    content.to_file(tmp_path / "content2.yaml", minimal=False)
    # minimal should produce min-set (input dict), but does not work currently


def test_base_model_shepherd_fault_load_other(tmp_path: Path) -> None:
    path = tmp_path / "content.yaml"
    content = ContentModel(name="tricky", owner="peter", group="work")
    content.to_file(path)
    with pytest.raises(ValueError):  # noqa: PT011
        CalibrationCape.from_file(path)


def test_base_model_shepherd_fault_mutation() -> None:
    content = ContentModel(name="tricky", owner="peter", group="work")
    with pytest.raises(ValidationError):
        content.name = "whatever"


def test_base_model_shepherd_fault_short_str() -> None:
    with pytest.raises(ValidationError):
        _ = ContentModel(name="", owner="peter", group="work")


def test_base_model_shepherd_fault_long_str() -> None:
    with pytest.raises(ValidationError):
        _ = ContentModel(
            name="very_long_123456789_123456789_1234567890", owner="peter", group="work"
        )


def test_base_model_shepherd_fault_unsafe_str() -> None:
    with pytest.raises(ValidationError):
        _ = ContentModel(name="fs_trouble<>", owner="peter", group="work")
    with pytest.raises(ValidationError):
        _ = ContentModel(name="unicode", owner="peter", group="work", description="🐍")



"""Wrapper-related ecosystem for transferring models."""

from datetime import datetime
from typing import Annotated

from pydantic import BaseModel
from pydantic import StringConstraints

from shepherd_core.version import version

SafeStrClone = Annotated[str, StringConstraints(pattern=r"^[ -~]+$")]
# ⤷ copy avoids circular import


class Wrapper(BaseModel):
    """Generalized web- & file-interface for all models with dynamic typecasting."""

    datatype: str
    """ ⤷ model-name"""
    comment: SafeStrClone | None = None
    created: datetime | None = None
    """ ⤷ Optional metadata"""
    lib_ver: str | None = version
    """ ⤷ for debug-purposes and later compatibility-checks"""
    parameters: dict
    """ ⤷ ShpModel"""



"""Python related inventory model."""

import platform
from contextlib import suppress
from importlib import import_module

from pydantic import ConfigDict
from typing_extensions import Self

from shepherd_core.data_models import ShpModel


class PythonInventory(ShpModel):
    """Python related inventory model."""

    # program versions
    h5py: str | None = None
    numpy: str | None = None
    pydantic: str | None = None
    python: str | None = None
    shepherd_core: str | None = None
    shepherd_sheep: str | None = None
    yaml: str | None = None
    zstandard: str | None = None

    model_config = ConfigDict(str_min_length=0)

    @classmethod
    def collect(cls) -> Self:
        model_dict = {"python": platform.python_version()}
        module_names = [
            "h5py",
            "numpy",
            "pydantic",
            "shepherd_core",
            "shepherd_sheep",
            "yaml",
            "zstandard",
        ]

        for module_name in module_names:
            with suppress(ImportError):
                module = import_module(module_name)
                model_dict[module_name] = module.__version__
                globals()

        return cls(**model_dict)



"""Model for user-management."""

import secrets
from hashlib import pbkdf2_hmac
from typing import Annotated
from typing import Any

from pydantic import EmailStr
from pydantic import Field
from pydantic import SecretBytes
from pydantic import SecretStr
from pydantic import StringConstraints
from pydantic import model_validator
from pydantic import validate_call

from shepherd_core.data_models.base.content import NameStr
from shepherd_core.data_models.base.content import SafeStr
from shepherd_core.data_models.base.content import id_default
from shepherd_core.data_models.base.shepherd import ShpModel


@validate_call
def hash_password(pw: Annotated[str, StringConstraints(min_length=20, max_length=100)]) -> bytes:
    """Generate a hash of a string.

    # NOTE: 1M Iterations need 25s on beaglebone
    # TODO: add salt of testbed -> this fn should be part of Testbed-Object
    """
    return pbkdf2_hmac(
        "sha512",
        password=pw.encode("utf-8"),
        salt=b"testbed_salt_TODO",
        iterations=1_000_000,
        dklen=128,
    )


class User(ShpModel):
    """meta-data representation of a testbed-component (physical object)."""

    id: int = Field(
        description="Unique ID",
        default_factory=id_default,
    )
    name: NameStr
    description: SafeStr | None = None
    comment: SafeStr | None = None

    name_full: NameStr | None = None
    group: NameStr
    email: EmailStr

    pw_hash: SecretBytes | None = None
    # ⤷ was hash_password("this_will_become_a_salted_slow_hash") -> slowed BBB down
    # ⤷ TODO (min_length=128, max_length=512)

    token: SecretStr
    # ⤷ TODO (min_length=128), request with: token.get_secret_value()
    active: bool = False

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        # TODO:

        # post correction
        if values.get("token") is None:
            values["token"] = "shepherd_token_" + secrets.token_urlsafe(nbytes=128)

        return values



"""Converter for ELF-files."""

import subprocess
from pathlib import Path

from pydantic import validate_call

# extra src-file necessary to prevent circular import


@validate_call
def elf_to_hex(file_elf: Path, file_hex: Path | None = None) -> Path:
    """Convert ELF to hex file using objcopy."""
    if not file_elf.is_file():
        raise ValueError("Fn needs an existing file as input")
    if not file_hex:
        file_hex = file_elf.resolve().with_suffix(".hex")
    cmd = ["objcopy", "-O", "ihex", file_elf.resolve().as_posix(), file_hex.as_posix()]
    # TODO: observe - maybe $ARCH-Versions of objcopy are needed
    #  (hex of nRF / msp identical between the 3 $arch-versions)
    try:
        subprocess.run(cmd, check=True)  # noqa: S603
    except FileNotFoundError as err:
        raise RuntimeError(
            "Objcopy not found -> are binutils or build-essential installed?"
        ) from err
    except subprocess.CalledProcessError:
        raise RuntimeError("Objcopy failed to convert ELF to iHEX") from None
    return file_hex



"""How-to for defining an experiment - variant 1.

What the code does:
- recommended approach for missing testbed-client
- variants
- var1 - references a server-path for the firmware
- var2 - embeds local firmware in yaml (elf-support is linux-only)
- assumption:
- start ASAP,
- no custom IDs,
- static Power-Supply
- no power-tracing

"""

from pathlib import Path

import shepherd_core.data_models as sm
from shepherd_core.data_models.task import TestbedTasks
from shepherd_core.data_models.testbed import MCU

from shepherd_core import WebClient

# For online-queries the lib can be connected to the testbed-server.
# NOTE: there are 3 states:
# - unconnected -> demo-fixtures are queried (locally)
# - connected -> publicly available data is queried online
# - logged in with valid token -> also private data is queried online
do_connect = False
if do_connect:
    WebClient()

xp = sm.Experiment(
    name="meaningful_TestName",
    # time_start could be "2033-03-13 14:15:16" or "datetime.now() + timedelta(minutes=30)"
    duration=30,
    target_configs=[
        sm.TargetConfig(
            target_IDs=range(7, 12),
            custom_IDs=range(1, 100),  # note: longer list is OK
            energy_env=sm.EnergyEnvironment(name="eenv_static_3000mV_50mA_3600s"),
            firmware1=sm.Firmware(
                name="FW_TestXYZ",
                data=Path("/var/shepherd/content/fw/nes_lab/nrf52_demo_rf/build.elf"),
                data_type=sm.FirmwareDType.path_elf,
                data_local=False,
                mcu=MCU(name="nRF52"),
            ),
            power_tracing=None,
            gpio_tracing=sm.GpioTracing(),
            uart_logging=sm.UartLogging(baudrate=115_200),
        ),
    ],
)
xp.to_file("experiment_generic_var1.yaml")

# Create a tasks-list for the testbed
tb_tasks = TestbedTasks.from_xp(xp)
tb_tasks.to_file("experiment_generic_var1_tbt.yaml")

# next steps:
# - copy to server:
#   scp ./experiment_generic_varX_tbt.yaml user@shepherd.cfaed.tu-dresden.de:/var/shepherd/content/
# - run with herd-tool:
#   shepherd-herd --verbose run --attach /var/shepherd/content/experiment_generic_varX_tbt.yaml



"""Bundled core features used by several systems.

Provides classes for storing and retrieving sampled IV data to/from
HDF5 files.

"""

from .data_models.base.calibration import Calc_t
from .data_models.base.calibration import CalibrationCape
from .data_models.base.calibration import CalibrationEmulator
from .data_models.base.calibration import CalibrationHarvester
from .data_models.base.calibration import CalibrationPair
from .data_models.base.calibration import CalibrationSeries
from .data_models.base.timezone import local_now
from .data_models.base.timezone import local_tz
from .data_models.task.emulation import Compression
from .inventory import Inventory
from .logger import get_verbose_level
from .logger import increase_verbose_level
from .logger import log
from .reader import Reader
from .testbed_client.client_web import WebClient
from .version import version
from .writer import Writer

__version__ = version

__all__ = [
    "Calc_t",
    "CalibrationCape",
    "CalibrationEmulator",
    "CalibrationHarvester",
    "CalibrationPair",
    "CalibrationSeries",
    "Compression",
    "Inventory",
    "Reader",
    "WebClient",
    "Writer",
    "get_verbose_level",
    "increase_verbose_level",
    "local_now",
    "local_tz",
    "log",
]



"""Data-model for recorded eEnvs."""

from enum import Enum
from pathlib import Path
from typing import Any

from pydantic import PositiveFloat
from pydantic import model_validator

from shepherd_core.data_models.base.content import ContentModel
from shepherd_core.testbed_client import tb_client


class EnergyDType(str, Enum):
    """Data-Type-Options for energy environments."""

    ivtrace = ivsample = ivsamples = "ivsample"
    ivsurface = ivcurve = ivcurves = "ivcurve"
    isc_voc = "isc_voc"


class EnergyEnvironment(ContentModel):
    """Recording of meta-data representation of a testbed-component."""

    # General Metadata & Ownership -> ContentModel

    data_path: Path
    data_type: EnergyDType
    data_local: bool = True
    """ ⤷ signals that file has to be copied to testbed"""

    duration: PositiveFloat
    energy_Ws: PositiveFloat
    valid: bool = False

    # TODO: scale up/down voltage/current
    # TODO: multiple files for one env
    # TODO: mean power as energy/duration
    # TODO: harvester, transducer

    # additional descriptive metadata, TODO: these are very solar-centered -> generalize
    light_source: str | None = None
    weather_conditions: str | None = None
    indoor: bool | None = None
    location: str | None = None

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        values, _ = tb_client.try_completing_model(cls.__name__, values)
        # TODO: figure out a way to crosscheck type with actual data
        return tb_client.fill_in_user_data(values)



"""Module for task-related data-modules.

These models import externally from all other model-modules!
"""

import pickle
from pathlib import Path

import yaml

from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.base.wrapper import Wrapper
from shepherd_core.logger import log

from .emulation import Compression
from .emulation import EmulationTask
from .firmware_mod import FirmwareModTask
from .harvest import HarvestTask
from .observer_tasks import ObserverTasks
from .programming import ProgrammingTask
from .testbed_tasks import TestbedTasks

__all__ = [
    "Compression",
    "EmulationTask",
    "FirmwareModTask",
    "HarvestTask",
    "ObserverTasks",
    "ProgrammingTask",
    "TestbedTasks",
    "extract_tasks",
    "prepare_task",
]


def prepare_task(config: ShpModel | Path | str, observer: str | None = None) -> Wrapper:
    """Open file and extract tasks.

    - Open file (from Path or str of Path)
    - wraps task-model
    - and if it's an TestbedTasks it will extract the correct ObserverTask
    """
    if isinstance(config, str):
        config = Path(config)

    if isinstance(config, Path) and config.exists() and config.suffix.lower() == ".pickle":
        with config.resolve().open("rb") as shp_file:
            shp_dict = pickle.load(shp_file)  # noqa: S301
        shp_wrap = Wrapper(**shp_dict)
    elif isinstance(config, Path) and config.exists() and config.suffix.lower() == ".yaml":
        with config.resolve().open() as shp_file:
            shp_dict = yaml.safe_load(shp_file)
        shp_wrap = Wrapper(**shp_dict)
    elif isinstance(config, ShpModel):
        shp_wrap = Wrapper(
            datatype=type(config).__name__,
            parameters=config.model_dump(),
        )
    else:
        msg = f"had unknown input: {type(config)}"
        raise TypeError(msg)

    if shp_wrap.datatype == TestbedTasks.__name__:
        if observer is None:
            log.debug(
                "Task-Set contained TestbedTasks & no observer was provided -> will return TB-Tasks"
            )
            return shp_wrap
        tbt = TestbedTasks(**shp_wrap.parameters)
        log.debug("Loading Testbed-Tasks %s for %s", tbt.name, observer)
        obt = tbt.get_observer_tasks(observer)
        if obt is None:
            msg = f"Observer '{observer}' is not in TestbedTask-Set"
            raise ValueError(msg)
        shp_wrap = Wrapper(
            datatype=type(obt).__name__,
            parameters=obt.model_dump(),
        )
    return shp_wrap


def extract_tasks(shp_wrap: Wrapper, *, no_task_sets: bool = True) -> list[ShpModel]:
    """Make the individual task-sets usable for each observer."""
    if shp_wrap.datatype == ObserverTasks.__name__:
        obt = ObserverTasks(**shp_wrap.parameters)
        content = obt.get_tasks()
    elif shp_wrap.datatype == EmulationTask.__name__:
        content = [EmulationTask(**shp_wrap.parameters)]
    elif shp_wrap.datatype == HarvestTask.__name__:
        content = [HarvestTask(**shp_wrap.parameters)]
    elif shp_wrap.datatype == FirmwareModTask.__name__:
        content = [FirmwareModTask(**shp_wrap.parameters)]
    elif shp_wrap.datatype == ProgrammingTask.__name__:
        content = [ProgrammingTask(**shp_wrap.parameters)]
    elif shp_wrap.datatype == TestbedTasks.__name__:
        if no_task_sets:
            raise ValueError("Model in Wrapper was TestbedTasks -> Task-Sets not allowed!")
        content = [TestbedTasks(**shp_wrap.parameters)]
    else:
        msg = f"Extractor had unknown task: {shp_wrap.datatype}"
        raise ValueError(msg)

    return content



from collections.abc import Callable

import pytest
from shepherd_core.calibration_hw_def import adc_current_to_raw
from shepherd_core.calibration_hw_def import adc_raw_to_current
from shepherd_core.calibration_hw_def import adc_raw_to_voltage
from shepherd_core.calibration_hw_def import adc_voltage_to_raw
from shepherd_core.calibration_hw_def import dac_raw_to_voltage
from shepherd_core.calibration_hw_def import dac_voltage_to_raw


@pytest.mark.parametrize("fn", [adc_current_to_raw, adc_voltage_to_raw, dac_voltage_to_raw])
def test_convert_to_raw(fn: Callable) -> None:
    values = [-1, -1e-6, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1e-0]
    value_prev = -1
    for value in values:
        value_raw = fn(value)
        assert value_raw >= 0
        assert value_raw < 2**18
        assert value_raw >= value_prev
        value_prev = value_raw


@pytest.mark.parametrize("fn", [adc_raw_to_current, adc_raw_to_voltage, dac_raw_to_voltage])
def test_convert_to_si(fn: Callable) -> None:
    values = [-100, -1, 0, 1, 2**6, 2**12, 2**18, 2**24]
    value_prev = -1
    for value in values:
        value_si = fn(value)
        assert value_si >= 0
        assert value_si <= 6
        assert value_si >= value_prev
        value_prev = value_si



import subprocess
from collections.abc import Iterable
from pathlib import Path

import numpy as np
import pytest

from shepherd_core import Writer


def generate_h5_file(file_path: Path, file_name: str = "harvest_example.h5") -> Path:
    store_path = file_path / file_name

    with Writer(store_path) as file:
        file.store_hostname("artificial")

        duration_s = 2
        repetitions = 5
        timestamp_vector = np.arange(0.0, duration_s, file.sample_interval_ns / 1e9)

        # values in SI units
        voltages = np.linspace(3.60, 1.90, int(file.samplerate_sps * duration_s))
        currents = np.linspace(100e-6, 2000e-6, int(file.samplerate_sps * duration_s))

        for idx in range(repetitions):
            timestamps = idx * duration_s + timestamp_vector
            file.append_iv_data_si(timestamps, voltages, currents)

    return store_path


@pytest.fixture
def data_h5(tmp_path: Path) -> Path:
    return generate_h5_file(tmp_path)


def pytest_collection_modifyitems(
    config: pytest.Config,
    items: Iterable[pytest.Item],
) -> None:
    # ELF
    try:
        from pwnlib.elf import ELF  # noqa: PLC0415
    except ImportError:
        ELF = None
    skip_elf = pytest.mark.skip(
        reason="ELF-support not found -> shepherd_core[elf] missing or OS is Windows?"
    )

    # OBJCOPY
    try:
        subprocess.run(["objcopy", "--version"], check=True)
        OBJCOPY = True
    except FileNotFoundError:
        OBJCOPY = None
    skip_converter = pytest.mark.skip(
        reason="Objcopy not found -> are binutils or build-essential installed?"
    )

    for item in items:
        if "elf" in item.keywords and ELF is None:
            item.add_marker(skip_elf)
        if "converter" in item.keywords and OBJCOPY is None:
            item.add_marker(skip_converter)



import sys
from pathlib import Path

import h5py
import pytest
import yaml
from pydantic import ValidationError

from shepherd_core import Reader
from shepherd_core import Writer


def test_reader_metadata(data_h5: Path) -> None:
    with Reader(data_h5, verbose=True) as sfr:
        assert sfr.energy() > 0
        assert sfr.is_valid()

        meta_data_a = sfr.save_metadata()
        meta_path = data_h5.resolve().with_suffix(".yaml")
        assert meta_path.exists()
        with meta_path.open() as meta_file:
            meta_data_b = yaml.safe_load(meta_file)
            assert len(meta_data_b) == len(meta_data_a)
            assert sys.getsizeof(meta_data_b) == sys.getsizeof(meta_data_a)


def test_reader_items(data_h5: Path) -> None:
    """Access internal"""
    with Reader(data_h5, verbose=True) as sfr:
        _ = sfr["data"]  # group
        _ = sfr["mode"]  # attribute
        with pytest.raises(KeyError):
            _ = sfr["non-existing"]


def test_reader_open(tmp_path: Path) -> None:
    with pytest.raises(ValidationError):
        Reader(file_path=None)

    tmp_file = (tmp_path / "data.h5").resolve()

    with pytest.raises(FileNotFoundError):
        Reader(file_path=tmp_file)

    with tmp_file.open("w") as tf:
        tf.write("abc def ghi")
    with pytest.raises(TypeError):
        Reader(file_path=tmp_file)


def test_reader_meta_repr(data_h5: Path) -> None:
    with Reader(data_h5, verbose=True) as sfr:
        print(sfr)
        print(sfr.get_calibration_data())
        print(sfr.get_hrv_config())


def test_reader_read_open_end(data_h5: Path) -> None:
    with Reader(data_h5, verbose=True) as sfr:
        for buf in sfr.read():
            print(len(buf[0]))


def test_reader_read_5(data_h5: Path) -> None:
    with Reader(data_h5, verbose=True) as sfr:
        for buf in sfr.read(end_n=5):
            print(len(buf[0]))


def test_reader_read_start_late(data_h5: Path) -> None:
    with Reader(data_h5, verbose=True) as sfr:
        for buf in sfr.read(start_n=2):
            print(len(buf[0]))


def test_reader_read_raw(data_h5: Path) -> None:
    with Reader(data_h5, verbose=True) as sfr:
        for buf in sfr.read(is_raw=True):
            print(len(buf[0]))


def test_reader_non_valid(tmp_path: Path) -> None:
    data_h5 = tmp_path / "data.h5"
    with h5py.File(data_h5, "w") as h5f:
        grp_data = h5f.create_group("data")
        grp_data.create_dataset("time", (0,))
        grp_data.create_dataset("current", (0,))
        grp_data.create_dataset("voltage", (0,))
        grp_data["time"].attrs["gain"] = 1
        grp_data["current"].attrs["gain"] = 1
        grp_data["voltage"].attrs["gain"] = 1
        grp_data["time"].attrs["offset"] = 0
        grp_data["current"].attrs["offset"] = 0
        grp_data["voltage"].attrs["offset"] = 0
    with Reader(data_h5, verbose=True) as sfr:
        # try getting metadata -> no exception but safe defaults
        sfr.get_window_samples()
        sfr.get_mode()
        sfr.get_mode()
        sfr.get_hostname()
        sfr.get_datatype()
        assert not sfr.is_valid()


def test_reader_fault_path(tmp_path: Path) -> None:
    with pytest.raises(FileNotFoundError):
        Reader(tmp_path, verbose=True)


def test_reader_fault_no_data(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        del sfw.h5file["data"]
    with pytest.raises(KeyError):
        Reader(data_h5, verbose=True)


def test_reader_fault_no_mode(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        del sfw.h5file.attrs["mode"]
    with Reader(data_h5, verbose=True) as sfr:
        assert not sfr.is_valid()


def test_reader_fault_fake_mode(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        sfw.h5file.attrs["mode"] = "burning"
    with Reader(data_h5, verbose=True) as sfr:
        assert not sfr.is_valid()


def test_reader_fault_no_window_samples(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        del sfw.h5file["data"].attrs["window_samples"]
    with Reader(data_h5, verbose=True) as sfr:
        assert not sfr.is_valid()


def test_reader_fault_no_datatype(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        del sfw.h5file["data"].attrs["datatype"]
    with Reader(data_h5, verbose=True) as sfr:
        assert not sfr.is_valid()


def test_reader_fault_wrong_datatype(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        sfw.h5file["data"].attrs["datatype"] = "apples"
    with Reader(data_h5, verbose=True) as sfr:
        assert not sfr.is_valid()


def test_reader_fault_wrong_window1(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        sfw["data"].attrs["datatype"] = "ivcurve"
        sfw["data"].attrs["window_samples"] = 0
    with Reader(data_h5, verbose=True) as sfr:
        assert not sfr.is_valid()


def test_reader_fault_wrong_window2(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        sfw["data"].attrs["datatype"] = "ivsample"
        sfw["data"].attrs["window_samples"] = 1111
    with Reader(data_h5, verbose=True) as sfr:
        # only triggers warning
        assert sfr.is_valid()


def test_reader_fault_no_current(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        del sfw.h5file["data"]["current"]
    with pytest.raises(KeyError):
        Reader(data_h5, verbose=True)


def test_reader_fault_no_voltage(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        del sfw.h5file["data"]["voltage"]
    with pytest.raises(KeyError):
        Reader(data_h5, verbose=True)


def test_reader_fault_non_eq_time(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        sfw.h5file["data"]["time"].resize((0,))
    with Reader(data_h5, verbose=True) as sfr:
        assert sfr.is_valid()
        # soft-criteria ... reader just complains


def test_reader_fault_unaligned(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        new_size = sfw.CHUNK_SAMPLES_N / 2
        sfw.h5file["data"]["time"].resize((new_size,))
        sfw.h5file["data"]["voltage"].resize((new_size,))
        sfw.h5file["data"]["current"].resize((new_size,))
    with Reader(data_h5, verbose=True) as sfr:
        assert sfr.is_valid()
        # soft-criteria ... reader just complains


def test_reader_fault_strange_compression(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        time = sfw.h5file["data"]["time"][:]
        del sfw.h5file["data"]["time"]
        sfw.h5file["data"].create_dataset(
            "time",
            (0,),
            dtype="u8",
            maxshape=(None,),
            compression="lzf",  # something like szip wanted
        )
        sfw.h5file["data"]["time"].resize(time.shape)
        sfw.h5file["data"]["time"][0 : time.shape[0]] = time[: time.shape[0]]
        sfw.h5file["data"]["time"].attrs["gain"] = 1
        sfw.h5file["data"]["time"].attrs["offset"] = 0
    with Reader(data_h5, verbose=True) as sfr:
        assert sfr.is_valid()
        # soft-criteria ... reader just complains


def test_reader_fault_slow_compression(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        time = sfw.h5file["data"]["time"][:]
        del sfw.h5file["data"]["time"]
        sfw.h5file["data"].create_dataset(
            "time",
            (0,),
            dtype="u8",
            maxshape=(None,),
            compression="gzip",
            compression_opts=9,
        )
        sfw.h5file["data"]["time"].resize(time.shape)
        sfw.h5file["data"]["time"][0 : time.shape[0]] = time[: time.shape[0]]
        sfw.h5file["data"]["time"].attrs["gain"] = 1
        sfw.h5file["data"]["time"].attrs["offset"] = 0
    with Reader(data_h5, verbose=True) as sfr:
        assert sfr.is_valid()
        # soft-criteria ... reader just complains


def test_reader_fault_no_hostname(data_h5: Path) -> None:
    with Writer(data_h5, modify_existing=True) as sfw:
        sfw.h5file.attrs["hostname"] = "unknown"
    with Reader(data_h5, verbose=True) as sfr:
        assert sfr.is_valid()
        # soft-criteria ... reader just complains


def test_reader_fault_jumps_timestamp(data_h5: Path) -> None:
    with Reader(data_h5, verbose=True) as sfr:
        assert sfr.check_timediffs()
    with Writer(data_h5, modify_existing=True) as sfw:
        sfw.h5file["data"]["time"][sfw.CHUNK_SAMPLES_N] = 0
    with Reader(data_h5, verbose=True) as sfr:
        assert not sfr.check_timediffs()


def test_reader_save_meta(data_h5: Path) -> None:
    with Reader(data_h5, verbose=True) as sfr:
        assert sfr.save_metadata() != {}
        # yaml now already exists
        assert sfr.save_metadata() == {}


def test_reader_save_meta_fail(data_h5: Path) -> None:
    with Reader(data_h5, verbose=True) as sfr:
        sfr.file_path = None
        assert sfr.save_metadata() == {}



"""System / OS related inventory model."""

import platform
import subprocess
import time
from collections.abc import Mapping
from collections.abc import Sequence
from datetime import datetime
from pathlib import Path
from types import MappingProxyType
from typing import Any

from typing_extensions import Self

from shepherd_core.data_models.base.timezone import local_now
from shepherd_core.logger import log

try:
    import psutil
except ImportError:
    psutil = None

from pydantic import ConfigDict
from pydantic import PositiveInt

from shepherd_core.data_models import ShpModel


class SystemInventory(ShpModel):
    """System / OS related inventory model."""

    uptime: PositiveInt
    """ ⤷ seconds"""
    timestamp: datetime

    system: str
    release: str
    version: str

    machine: str
    processor: str

    ptp: str | None = None

    hostname: str

    interfaces: Mapping[str, Any] = MappingProxyType({})
    # ⤷ tuple with
    #   ip IPvAnyAddress
    #   mac MACStr

    fs_root: Sequence[str] = ()
    beagle: Sequence[str] = ()

    model_config = ConfigDict(str_min_length=0)

    @classmethod
    def collect(cls) -> Self:
        ts = local_now()

        if psutil is None:
            ifs2 = {}
            uptime = 0
            log.warning(
                "Inventory-Parameters will be missing. "
                "Please install functionality with "
                "'pip install shepherd_core[inventory] -U' first"
            )
        else:
            ifs1 = psutil.net_if_addrs().items()
            ifs2 = {name: (_if[1].address, _if[0].address) for name, _if in ifs1 if len(_if) > 1}
            uptime = time.time() - psutil.boot_time()

        fs_cmd = ["/usr/bin/df", "-h", "/"]
        fs_out = []
        if Path(fs_cmd[0]).is_file():
            reply = subprocess.run(  # noqa: S603
                fs_cmd, timeout=30, capture_output=True, check=False
            )
            fs_out = str(reply.stdout).split(r"\n")

        beagle_cmd = ["/usr/bin/beagle-version"]
        beagle_out = []
        if Path(beagle_cmd[0]).is_file():
            reply = subprocess.run(  # noqa: S603
                beagle_cmd, timeout=30, capture_output=True, check=False
            )
            beagle_out = str(reply.stdout).split(r"\n")

        ptp_cmd = ["/usr/sbin/ptp4l", "-v"]
        ptp_out = None
        if Path(ptp_cmd[0]).is_file():
            reply = subprocess.run(  # noqa: S603
                ptp_cmd, timeout=30, capture_output=True, check=False
            )
            ptp_out = f"{reply.stdout}, {reply.stderr}"
            # alternative: check_output - seems to be lighter

        model_dict = {
            "uptime": round(uptime),
            "timestamp": ts,
            "system": platform.system(),
            "release": platform.release(),
            "version": platform.version(),
            "machine": platform.machine(),
            "processor": platform.processor(),
            "hostname": platform.node(),
            "interfaces": ifs2,
            "fs_root": fs_out,
            "beagle": beagle_out,
            "ptp": ptp_out,
        }

        return cls(**model_dict)



"""Log handler of shepherd."""

import logging
import logging.handlers

import chromalog

chromalog.basicConfig(format="%(message)s")
log = logging.getLogger("SHPCore")
log.addHandler(logging.NullHandler())

verbose_level: int = 2


def get_verbose_level() -> int:
    """Get log level of shepherd."""
    return verbose_level


def set_log_verbose_level(log_: logging.Logger | logging.Handler, verbose: int) -> None:
    """Set log level of shepherd."""
    if verbose == 0:
        log_.setLevel(logging.ERROR)
        logging.basicConfig(level=logging.ERROR)
    elif verbose == 1:
        log_.setLevel(logging.WARNING)
    elif verbose == 2:
        log_.setLevel(logging.INFO)
    elif verbose > 2:
        log_.setLevel(logging.DEBUG)

    if verbose < 3:
        # reduce log-overhead when not debugging, also more user-friendly exceptions
        logging._srcfile = None  # noqa: SLF001
        logging.logThreads = False
        logging.logProcesses = False

    if verbose > 2:
        chromalog.basicConfig(format="%(name)s %(levelname)s: %(message)s")
    else:
        chromalog.basicConfig(format="%(message)s")  # reduce internals


def increase_verbose_level(verbose: int) -> None:
    """Increase log level of shepherd."""
    global verbose_level  # noqa: PLW0603
    if verbose >= verbose_level:
        verbose_level = min(max(verbose, 0), 3)
        set_log_verbose_level(log, verbose_level)


increase_verbose_level(2)

# short reminder for format-strings:
# %s    string
# %d    decimal
# %f    float
# %o    decimal as octal
# %x    decimal as hex
#
# %05d  pad right (aligned with 5chars)
# %-05d pad left (left aligned)
# %06.2f    6chars float, including dec point, with 2 chars after
# %.5s  truncate to 5 chars
#
# %% for a percent character



"""Container for shepherds data-models.

Public models are directly referenced here and are usable like:

'''python
from shepherd_core import data_models

cdata = data_models.CapeData(serial_number="A123")
'''
"""

from .base.calibration import CalibrationCape
from .base.calibration import CalibrationEmulator
from .base.calibration import CalibrationHarvester
from .base.calibration import CalibrationPair
from .base.calibration import CalibrationSeries
from .base.calibration import CapeData
from .base.content import ContentModel
from .base.shepherd import ShpModel
from .base.wrapper import Wrapper
from .content.energy_environment import EnergyDType
from .content.energy_environment import EnergyEnvironment
from .content.firmware import Firmware
from .content.firmware import FirmwareDType
from .content.virtual_harvester_config import VirtualHarvesterConfig
from .content.virtual_source_config import VirtualSourceConfig
from .content.virtual_storage_config import VirtualStorageConfig
from .experiment.experiment import Experiment
from .experiment.observer_features import GpioActuation
from .experiment.observer_features import GpioEvent
from .experiment.observer_features import GpioLevel
from .experiment.observer_features import GpioTracing
from .experiment.observer_features import PowerTracing
from .experiment.observer_features import SystemLogging
from .experiment.observer_features import UartLogging
from .experiment.target_config import TargetConfig

__all__ = [
    "CalibrationCape",
    "CalibrationEmulator",
    "CalibrationHarvester",
    "CalibrationPair",
    "CalibrationSeries",
    "CapeData",
    "ContentModel",
    "EnergyDType",
    "EnergyEnvironment",
    "Experiment",
    "Firmware",
    "FirmwareDType",
    "GpioActuation",
    "GpioEvent",
    "GpioLevel",
    "GpioTracing",
    "PowerTracing",
    "ShpModel",
    "SystemLogging",
    "TargetConfig",
    "UartLogging",
    "VirtualHarvesterConfig",
    "VirtualSourceConfig",
    "VirtualStorageConfig",
    "Wrapper",
]



"""Current implementation of a file-based database."""

import copy
import pickle
from collections.abc import Iterable
from collections.abc import Mapping
from datetime import datetime
from datetime import timedelta
from pathlib import Path
from typing import Any

import yaml
from pydantic import validate_call
from typing_extensions import Self

from shepherd_core.data_models.base.timezone import local_now
from shepherd_core.data_models.base.timezone import local_tz
from shepherd_core.data_models.base.wrapper import Wrapper
from shepherd_core.logger import log

from .cache_path import cache_user_path

# Proposed field-name:
# - inheritance
# - inherit_from  <-- current choice
# - inheritor
# - hereditary
# - based_on


class Fixture:
    """Current implementation of a file-based database."""

    def __init__(self, model_type: str) -> None:
        self.model_type: str = model_type.lower()
        self.elements_by_name: dict[str, dict[str, Any]] = {}
        self.elements_by_id: dict[int, dict[str, Any]] = {}
        # Iterator reset
        self._iter_index: int = 0
        self._iter_list: list[dict[str, Any]] = list(self.elements_by_name.values())

    def insert(self, data: Wrapper) -> None:
        # ⤷ TODO: could get easier
        #    - when not model_name but class used
        #    - use doubleref name->id->data (saves RAM)
        if data.datatype.lower() != self.model_type.lower():
            return
        if "name" not in data.parameters:
            return
        name = str(data.parameters["name"]).lower()
        _id = data.parameters["id"]
        data_model = data.parameters
        self.elements_by_name[name] = data_model
        self.elements_by_id[_id] = data_model
        # update iterator
        self._iter_list: list[dict[str, Any]] = list(self.elements_by_name.values())

    def __getitem__(self, key: str | int) -> dict[str, Any]:
        original_key = key
        if isinstance(key, str):
            key = key.lower()
            if key in self.elements_by_name:
                return self.elements_by_name[key]
            if key.isdigit():
                key = int(key)
        if key in self.elements_by_id:
            return self.elements_by_id[int(key)]
        msg = f"{self.model_type} '{original_key}' not found!"
        raise ValueError(msg)

    def __iter__(self) -> Self:
        self._iter_index = 0
        self._iter_list = list(self.elements_by_name.values())
        return self

    def __next__(self) -> dict[str, Any]:
        if self._iter_index < len(self._iter_list):
            member = self._iter_list[self._iter_index]
            self._iter_index += 1
            return member
        raise StopIteration

    def keys(self) -> Iterable[str]:
        return self.elements_by_name.keys()

    def refs(self) -> dict:
        return {_i["id"]: _i["name"] for _i in self.elements_by_id.values()}

    def inheritance(
        self, values: dict[str, Any], chain: list[str] | None = None
    ) -> tuple[dict[str, Any], list[str]]:
        if chain is None:
            chain = []
        values = copy.copy(values)
        post_process: bool = False
        fixture_base: dict = {}
        if "inherit_from" in values:
            fixture_name = values.pop("inherit_from")
            # ⤷ will also remove entry from dict
            if "name" in values and len(chain) < 1:
                base_name = values.get("name")
                if base_name in chain:
                    msg = f"Inheritance-Circle detected ({base_name} already in {chain})"
                    raise ValueError(msg)
                if base_name == fixture_name:
                    msg = f"Inheritance-Circle detected ({base_name} == {fixture_name})"
                    raise ValueError(msg)
                chain.append(base_name)
            fixture_base = copy.copy(self[fixture_name])
            log.debug("'%s' will inherit from '%s'", self.model_type, fixture_name)
            fixture_base["name"] = fixture_name
            chain.append(fixture_name)
            base_dict, chain = self.inheritance(values=fixture_base, chain=chain)
            for key, value in values.items():
                # keep previous entries
                base_dict[key] = value
            values = base_dict

        # TODO: cleanup and simplify - use fill_mode() and line up with web-interface
        elif "name" in values and values.get("name").lower() in self.elements_by_name:
            fixture_name = values.get("name").lower()
            fixture_base = copy.copy(self.elements_by_name[fixture_name])
            post_process = True

        elif values.get("id") in self.elements_by_id:
            _id = values["id"]
            fixture_base = copy.copy(self.elements_by_id[_id])
            post_process = True

        if post_process:
            # last two cases need
            for key, value in values.items():
                # keep previous entries
                fixture_base[key] = value
            if "inherit_from" in fixture_base:
                # as long as this key is present this will act recursively
                chain.append(fixture_base["name"])
                values, chain = self.inheritance(values=fixture_base, chain=chain)
            else:
                values = fixture_base

        return values, chain

    @staticmethod
    def fill_model(model: Mapping, base: dict) -> dict:
        base = copy.copy(base)
        for key, value in model.items():
            # keep previous entries
            base[key] = value
        return base

    def query_id(self, _id: int) -> dict[str, Any]:
        if isinstance(_id, int) and _id in self.elements_by_id:
            return self.elements_by_id[_id]
        msg = f"Initialization of {self.model_type} by ID failed - {_id} is unknown!"
        raise ValueError(msg)

    def query_name(self, name: str) -> dict[str, Any]:
        if isinstance(name, str) and name.lower() in self.elements_by_name:
            return self.elements_by_name[name.lower()]
        msg = f"Initialization of {self.model_type} by name failed - {name} is unknown!"
        raise ValueError(msg)


def file_older_than(file: Path, delta: timedelta) -> bool:
    """Decide if file is older than a specific duration of time."""
    cutoff = local_now() - delta
    mtime = datetime.fromtimestamp(file.stat().st_mtime, tz=local_tz())
    return mtime < cutoff


class Fixtures:
    """A collection of individual fixture-elements."""

    suffix = ".yaml"

    @validate_call
    def __init__(self, file_path: Path | None = None, *, reset: bool = False) -> None:
        if file_path is None:
            self.file_path = Path(__file__).parent.parent.resolve() / "data_models"
        else:
            self.file_path = file_path
        self.components: dict[str, Fixture] = {}
        cache_file = cache_user_path / "fixtures.pickle"
        sheep_detect = Path("/lib/firmware/am335x-pru0-fw").exists()

        if (
            not sheep_detect
            and cache_file.exists()
            and not file_older_than(cache_file, timedelta(hours=24))
            and not reset
        ):
            # speedup by loading from cache
            # TODO: also add version as criterion
            with cache_file.open("rb", buffering=-1) as fd:
                self.components = pickle.load(fd)  # noqa: S301
            log.debug(" -> found & used pickled fixtures")
        else:
            if self.file_path.is_file():
                files = [self.file_path]
            elif self.file_path.is_dir():
                files = list(
                    self.file_path.glob("**/*" + self.suffix)
                )  # for py>=3.12: case_sensitive=False
                log.debug(" -> got %s %s-files", len(files), self.suffix)
            else:
                raise ValueError("Path must either be file or directory (or empty)")

            for file in files:
                self.insert_file(file)

            if len(self.components) < 1:
                log.error(f"No fixture-components found at {self.file_path.as_posix()}")
            elif sheep_detect:
                cache_file.parent.mkdir(parents=True, exist_ok=True)
                with cache_file.open("wb", buffering=-1) as fd:
                    pickle.dump(self.components, fd)

    @validate_call
    def insert_file(self, file: Path) -> None:
        with file.open() as fd:
            fixtures = yaml.safe_load(fd)
            for fixture in fixtures:
                if not isinstance(fixture, dict):
                    continue
                fix_wrap = Wrapper(**fixture)
                self.insert_model(fix_wrap)

    def insert_model(self, data: Wrapper) -> None:
        fix_type = data.datatype.lower()
        if self.components.get(fix_type) is None:
            self.components[fix_type] = Fixture(model_type=fix_type)
        self.components[fix_type].insert(data)

    def __getitem__(self, key: str) -> Fixture:
        key = key.lower()
        if key in self.components:
            return self.components[key]
        msg = f"Component '{key}' not found!"
        raise ValueError(msg)

    def keys(self) -> Iterable[str]:
        return self.components.keys()

    @staticmethod
    def to_file(file: Path) -> None:
        msg = f"TODO (val={file})"
        raise NotImplementedError(msg)



from pathlib import Path

import pytest
from shepherd_core.data_models import EnergyDType
from shepherd_core.data_models import VirtualSourceConfig
from shepherd_core.vsource import VirtualSourceModel

from shepherd_core import CalibrationEmulator
from shepherd_core import Reader

# virtual_converter_model gets tested below with vsrc_model

src_list = [
    "direct",
    "diode+capacitor",
    "diode+resistor+capacitor",
    "BQ25504",
    "BQ25504s",
    "BQ25570",
    "BQ25570s",
]


def src_model(
    name: str,
    dtype_in: EnergyDType = EnergyDType.ivsample,
    window_size: int | None = None,
    voltage_step_V: float | None = None,
) -> VirtualSourceModel:
    src_config = VirtualSourceConfig(name=name)
    cal_emu = CalibrationEmulator()
    return VirtualSourceModel(
        src_config,
        cal_emu,
        log_intermediate=False,
        dtype_in=dtype_in,
        window_size=window_size,
        voltage_step_V=voltage_step_V,
    )


@pytest.mark.parametrize("src_name", src_list)
def test_vsource_vsrc_min(src_name: str) -> None:
    src = src_model(src_name)
    src.iterate_sampling()


def test_vsource_vsrc_static1() -> None:
    iterations = 2_000
    src = src_model("BQ25504")
    for _ in range(iterations):
        src.iterate_sampling(V_inp_uV=3_000_000, I_inp_nA=0)
    assert src.W_inp_fWs == 0.0
    assert src.W_out_fWs == 0.0  # -> leakage now locally in storage
    # pytest.approx(c_leak_fWs(src, iterations), rel=1e-4, abs=1e-6))


def test_vsource_vsrc_static2() -> None:
    iterations = 2_000
    src = src_model("BQ25504")
    for _ in range(iterations):
        src.iterate_sampling(V_inp_uV=0, I_inp_nA=3_000_000)
    assert src.W_inp_fWs == 0.0
    assert src.W_out_fWs == 0.0  # -> leakage now locally in storage


@pytest.mark.parametrize("src_name", src_list[2:])
def test_vsource_charge(src_name: str) -> None:
    iterations = 8_000
    src = src_model(src_name)
    for v_mV in range(iterations):
        src.iterate_sampling(V_inp_uV=10**6 + v_mV * 1000, I_inp_nA=1_500_000)
    v_out = src.iterate_sampling(V_inp_uV=1_000_000, I_inp_nA=1_000_000)
    assert src.W_inp_fWs > 0.0
    assert src.W_out_fWs == 0.0  # -> leakage now locally in storage
    assert v_out > 0.0


@pytest.mark.parametrize("src_name", src_list[4:])
def test_vsource_drain(src_name: str) -> None:
    iterations = 4_000
    src = src_model(src_name)
    assert src.W_inp_fWs == 0.0
    # pre-charge and then drain
    for v_mV in range(iterations):
        src.iterate_sampling(V_inp_uV=v_mV * 1000, I_inp_nA=2_000_000)
    src.W_inp_fWs = 0.0
    for c_uA in range(iterations):
        src.iterate_sampling(I_out_nA=c_uA * 1000)
    v_out = src.iterate_sampling()
    assert src.W_inp_fWs == 0.0
    assert src.W_out_fWs > 0.0  # -> leakage now locally in storage
    assert v_out >= 0.0


def test_vsource_vsrc_over_voltage() -> None:
    iterations = 100
    src = src_model("BQ25504")
    for _ in range(iterations):
        src.iterate_sampling(V_inp_uV=10 * 10**6, I_inp_nA=3_000_000)
    assert src.cnv.V_input_uV <= 5 * 10**6
    assert src.W_inp_fWs > 0.0


def test_vsource_vsrc_over_current() -> None:
    iterations = 100
    src = src_model("BQ25504")
    for _ in range(iterations):
        src.iterate_sampling(V_inp_uV=5 * 10**6, I_inp_nA=100 * 10**6)
    assert src.W_inp_fWs > 0.0


def test_vsource_vsrc_cycle() -> None:
    iterations = 2000
    src = src_model("BQ25504s")

    for _ in range(iterations):
        src.iterate_sampling(V_inp_uV=5 * 10**6, I_inp_nA=4 * 10**6)
    v_out = src.iterate_sampling()
    assert v_out > 0

    for _ in range(iterations):
        src.iterate_sampling(I_out_nA=40 * 10**6)
    v_out = src.iterate_sampling()
    assert v_out == 0
    # TODO: not accurate anymore as the output does not get disconnected for this BQ

    for _ in range(iterations):
        src.iterate_sampling(V_inp_uV=5 * 10**6, I_inp_nA=20 * 10**6)
    v_out = src.iterate_sampling()
    assert v_out > 0

    assert src.W_out_fWs > 0.0  # -> leakage now locally in storage
    assert src.W_inp_fWs > src.W_out_fWs


def test_vsource_vsrc_create_files(
    file_ivcurve: Path, file_ivsample: Path, file_isc_voc: Path
) -> None:
    pass


@pytest.mark.parametrize("src_name", src_list)
def test_vsource_vsrc_sim_curve(src_name: str, file_ivcurve: Path) -> None:
    with Reader(file_ivcurve) as file:
        src = src_model(
            "BQ25504s",
            dtype_in=file.get_datatype(),
            window_size=file.get_window_samples(),
            voltage_step_V=file.get_voltage_step(),
        )
        for _t, _v, _i in file.read():
            length = max(_v.size, _i.size)
            for _n in range(length):
                src.iterate_sampling(V_inp_uV=_v[_n] * 10**6, I_inp_nA=_i[_n] * 10**9)


@pytest.mark.parametrize("src_name", src_list)
def test_vsource_vsrc_sim_sample(src_name: str, file_ivsample: Path) -> None:
    with Reader(file_ivsample) as file:
        src = src_model(
            "BQ25504s",
            dtype_in=file.get_datatype(),
            window_size=file.get_window_samples(),
            voltage_step_V=file.get_voltage_step(),
        )
        for _t, _v, _i in file.read():
            length = max(_v.size, _i.size)
            for _n in range(length):
                src.iterate_sampling(V_inp_uV=_v[_n] * 10**6, I_inp_nA=_i[_n] * 10**9)



"""Playground to determine the best integer-math and fit for constants."""

from itertools import product

from shepherd_core import log
from shepherd_core.data_models.content.virtual_storage_config import LuT_SIZE_LOG


def u32l(i: float) -> int:
    """Guard to supervise calculated model-states."""
    if i >= 2**32:
        log.warning("u32-overflow (%d)", i)
    if i < 0:
        log.warning("u32-underflow (%d)", i)
    return round(min(max(i, 0), 2**32 - 1))


# #### I_charge-to-dSoC #####
# Goal: Allow ~1 uF Capacitor to 800 mAh battery

dt_s = 10e-6
qs_As = [1e-6 * 1 * 2.0, 1e-6 * 5 * 2.5, 1e-6 * 10 * 3.6, 800 * 3.6]
Constant_1u_per_nA_n40 = [u32l((2**40 / 1e3) * dt_s / q_As) for q_As in qs_As]
Constant_1_per_nA_n60 = [u32l((2**60 / 1e9) * dt_s / q_As) for q_As in qs_As]

# #### SoC-to-position #####
# Goal: As-simple-as-possible

SoCs_1u_n32 = [round(x / 10 * 1e6 * 2**32) for x in range(11)]
LUT_SIZE = 128

# First Approach (1 Division in pos-calc -> impossible for PRU)
SoC_min_1u = round(1e6 / LUT_SIZE)
positions1 = [int(SoC_1u_n32 / SoC_min_1u / 2**32) for SoC_1u_n32 in SoCs_1u_n32]

# Second Approach
SoC_min = 1.0 / LUT_SIZE
inv_SoC_min_1M_n32 = round(2**32 / 1e6 / SoC_min)  # 1M / SoC_min
positions2 = [
    int(int(SoC_1u_n32 / 2**32) * inv_SoC_min_1M_n32 / 2**32) for SoC_1u_n32 in SoCs_1u_n32
]

# third approach
SoCs_1_n62 = [round(x / 10 * 2**62) for x in range(11)]
positions3 = [int(int(SoC_1_n62 / 2**32) * LUT_SIZE / 2**30) for SoC_1_n62 in SoCs_1_n62]

# final approach (upper u32 & rest of shift)
positions4 = [u32l(SoC_1_n62 // 2 ** (62 - LuT_SIZE_LOG)) for SoC_1_n62 in SoCs_1_n62]

# #### R_Leak-to-dSoC #####
# Goal: biggest possible dynamic range

Rs_leak_Ohm = [1e3, 10e3, 100e3, 1e6, 10e6]
Constants_1_per_V = [dt_s / q_As / R_leak_Ohm for q_As, R_leak_Ohm in product(qs_As, Rs_leak_Ohm)]
Constants_1u_per_uV_n40 = [u32l((2**40) * c_1_per_V) for c_1_per_V in Constants_1_per_V]
Constants_1_per_uV_n60 = [u32l((2**60 / 1e6) * c_1_per_V) for c_1_per_V in Constants_1_per_V]

print("done")  # noqa: T201



"""Generalized virtual energy storage data models (config).

Additions in near future:
- TODO: DC Bias to improve Capacitor-behavior, p_VOC / LuT

Possible Extensions:
- scale cell-count (determine how to change parameters) as
    2 cell Lipo or 2 cell lead would be advantageous
- add temperature-component?

"""

import math
import sys
from collections.abc import Sequence
from datetime import timedelta
from typing import Annotated
from typing import Any

from annotated_types import Ge
from annotated_types import Gt
from annotated_types import Le
from pydantic import Field
from pydantic import NonNegativeFloat
from pydantic import PositiveFloat
from pydantic import model_validator
from pydantic import validate_call
from typing_extensions import Self

from shepherd_core.config import config
from shepherd_core.data_models.base.content import ContentModel
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.logger import log
from shepherd_core.testbed_client import tb_client

soc_t = Annotated[float, Ge(0.0), Le(1.0)]
# TODO: adapt V_max in vsrc,
# TODO: do we need to set initial voltage, or is SoC ok? add V_OC_to_SoC()


class VirtualStorageConfig(ContentModel, title="Config for the virtual energy storage"):
    """KiBaM Battery model based on two papers.

    Model An Accurate Electrical Battery Model Capable
        of Predicting Runtime and I-V Performance
    https://rincon-mora.gatech.edu/publicat/jrnls/tec05_batt_mdl.pdf

    A Hybrid Battery Model Capable of Capturing Dynamic Circuit
        Characteristics and Nonlinear Capacity Effects
    https://digitalcommons.unl.edu/cgi/viewcontent.cgi?article=1210&context=electricalengineeringfacpub
    """

    SoC_init: soc_t = 0.8
    """ ⤷ State of Charge that is available when emulation starts.
    Allows a proper / fast startup.
    """

    q_As: PositiveFloat
    """ ⤷ Capacity (electrical charge) of Storage."""
    p_VOC: Annotated[Sequence[float], Field(min_length=6, max_length=6)] = [0, 0, 0, 1, 0, 0]
    """ ⤷ Parameters for V_OC-Mapping
        - direct SOC-Mapping by default
        - named a0 to a5 in paper
        """
    p_Rs: Annotated[Sequence[float], Field(min_length=6, max_length=6)] = [0, 0, 0, 0, 0, 0]
    """ ⤷ Parameters for series-resistance
        - no resistance set by default
        - named b0 to b5 in paper
        """
    p_RtS: Annotated[Sequence[float], Field(min_length=3, max_length=3)] = [0, 0, 0]
    """ ⤷ Parameters for R_transient_S (short-term),
        - no transient active by default
        - named c0 to c2 in paper
        """
    p_CtS: Annotated[Sequence[float], Field(min_length=3, max_length=3)] = [0, 0, 0]
    """ ⤷ Parameters for C_transient_S (short-term)
        - no transient active by default
        - named d0 to d2 in paper
        """
    p_RtL: Annotated[Sequence[float], Field(min_length=3, max_length=3)] = [0, 0, 0]
    """ ⤷ Parameters for R_transient_L (long-term)
        - no transient active by default
        - named e0 to e2 in paper
        """
    p_CtL: Annotated[Sequence[float], Field(min_length=3, max_length=3)] = [0, 0, 0]
    """ ⤷ Parameters for C_transient_L (long-term)
        - no transient active by default
        - named f0 to f2 in paper
        """

    p_rce: Annotated[float, Gt(0), Le(1.0)] = 1.0
    """ ⤷ Parameter for rate capacity effect
        - Set to 1 to disregard
        - named c in paper
        """
    kdash: PositiveFloat = sys.float_info.min  # TODO: use k directly?
    """ ⤷ Parameter for rate capacity effect
        - temporary component of rate capacity effect, valve in KiBaM (eq 17)
        - k' = k/c(1-c),
        """

    R_leak_Ohm: PositiveFloat = sys.float_info.max
    """ ⤷ Parameter for self discharge (custom extension)
        - effect is often very small, mostly relevant for some capacitors
        """

    @classmethod
    @validate_call
    def lipo(
        cls,
        q_mAh: PositiveFloat,  # TODO: charge is more correct
        SoC_init: soc_t | None = None,
        name: str | None = None,
        description: str | None = None,
    ) -> Self:
        """Modeled after the PL-383562 2C Polymer Lithium-ion Battery.

        Nominal Voltage     3.7 V
        Nominal Capacity    860 mAh
        Discharge Cutoff    3.0 V
        Charge Cutoff       4.2 V
        Max Discharge       2 C / 1.72 A
        https://www.batteryspace.com/prod-specs/pl383562.pdf

        """
        name_lmnts: list[str] = ["LiPo", f"{q_mAh:.0f}mAh", "3.7V"]
        if name is not None:
            name_lmnts = [name]  # specific name overrules params
        if description is None:
            description = "Model of a LiPo battery (3 to 4.2 V) with adjustable capacity"
        return cls(
            SoC_init=SoC_init if SoC_init else cls.model_fields["SoC_init"].default,
            q_As=q_mAh * 3600 / 1000,
            p_VOC=[-0.852, 63.867, 3.6297, 0.559, 0.51, 0.508],
            p_Rs=[0.1463, 30.27, 0.1037, 0.0584, 0.1747, 0.1288],
            p_RtS=[0.1063, 62.49, 0.0437],
            p_CtS=[-200, 138, 300],  # most likely a mistake (d1=-138) in the table/paper!
            p_RtL=[0.0712, 61.4, 0.0288],
            p_CtL=[-3083, 180, 5088],
            # y10 = 2863.3, y20 = 232.66 # unused
            p_rce=0.9248,
            kdash=0.0008,
            R_leak_Ohm=55.9e6 / q_mAh,  # from wiki ~ 5 % discharge/month
            # content-fields below
            name="_".join(name_lmnts),
            description=description,
            owner="NES Lab",
            group="NES Lab",
            visible2group=True,
            visible2all=True,
        )

    @classmethod
    @validate_call
    def lead_acid(
        cls,
        q_mAh: PositiveFloat,
        SoC_init: soc_t | None = None,
        name: str | None = None,
        description: str | None = None,
    ) -> Self:
        """Modeled after the LEOCH LP12-1.2AH lead acid battery.

        Nominal Voltage     12 V (6 Cell)
        Nominal Capacity    1.2 Ah
        Discharge Cutoff    10.8 V
        Charge Cutoff       13.5 V
        Max Discharge       15 C / 18 A
        https://www.leoch.com/pdf/reserve-power/agm-vrla/lp-general/LP12-1.2.pdf
        # NOTE: 1 cell has 2.1 V nom, cell-count as param?
        # NOTE: add temperature-component? -5mV/cell/K, also capacity-decrease
        """
        name_lmnts: list[str] = ["Lead-Acid", f"{q_mAh:.0f}mAh", "12V"]
        if name is not None:
            name_lmnts = [name]  # specific name overrules params
        if description is None:
            description = "Model of a 12V lead acid battery with adjustable capacity"
        return cls(
            SoC_init=SoC_init if SoC_init else cls.model_fields["SoC_init"].default,
            q_As=q_mAh * 3600 / 1000,
            p_VOC=[5.429, 117.5, 11.32, 2.706, 2.04, 1.026],
            p_Rs=[1.578, 8.527, 0.7808, -1.887, -2.404, -0.649],
            p_RtS=[2.771, 9.079, 0.22],
            p_CtS=[-2423, 75.14, 55],
            p_RtL=[2.771, 9.079, 0.218],
            # ⤷ first 2 values of p_RtL are identical with p_RtS in table/paper
            #   (strange, but plots look fine)
            p_CtL=[-1240, 9.571, 3100],
            # y10 = 2592, y20 = 1728 # unused
            p_rce=0.6,
            kdash=0.0034,
            R_leak_Ohm=174e6 / q_mAh,
            # ⤷ from datasheet - 3-20 % discharge/month for 1.2 Ah, here 5%
            # content-fields below
            name="_".join(name_lmnts),
            description=description,
            owner="NES Lab",
            group="NES Lab",
            visible2group=True,
            visible2all=True,
        )

    @classmethod
    @validate_call
    def capacitor(
        cls,
        C_uF: PositiveFloat,
        V_rated: PositiveFloat,
        SoC_init: soc_t | None = None,
        R_series_Ohm: NonNegativeFloat | None = None,
        R_leak_Ohm: PositiveFloat | None = None,
        name: str | None = None,
        description: str | None = None,
    ) -> Self:
        name_lmnts: list[str] = ["Capacitor", f"{C_uF:.0f}uF", f"{V_rated:.1f}V"]
        if name is not None:
            name_lmnts = [name]  # specific name overrules params
        if description is None:
            description = "Model of an Capacitor with various effects"
        if R_leak_Ohm is not None:
            description += f", R_leak = {R_leak_Ohm / 1e3:.3f} Ohm"
        if R_series_Ohm is not None:
            description += f", R_serial = {R_series_Ohm:.0f} Ohm"
        return cls(
            SoC_init=SoC_init if SoC_init else cls.model_fields["SoC_init"].default,
            q_As=1e-6 * C_uF * V_rated,
            p_VOC=[0, 0, 0, V_rated, 0, 0],  # 100% SoC is @V_rated,
            # no transients per default
            p_Rs=[0, 0, R_series_Ohm, 0, 0, 0]
            if R_series_Ohm
            else cls.model_fields["p_Rs"].default,  # const series resistance
            R_leak_Ohm=R_leak_Ohm if R_leak_Ohm else cls.model_fields["R_leak_Ohm"].default,
            # content-fields below
            name="_".join(name_lmnts),
            description=description,
            owner="NES Lab",
            group="NES Lab",
            visible2group=True,
            visible2all=True,
        )

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        values, chain = tb_client.try_completing_model(cls.__name__, values)
        values = tb_client.fill_in_user_data(values)
        log.debug("vStorage-Inheritances: %s", chain)
        return values

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        return self

    def without_rate_capacity(self) -> Self:
        model_dict = self.model_dump()
        model_dict["p_rce"] = 1
        model_dict["name"] += " no_rate_cap"
        return type(self)(**model_dict)

    def without_transient_voltages(self) -> Self:
        model_dict = self.model_dump()
        model_dict["p_RtS"] = [0, 0, 0]
        model_dict["p_CtS"] = [0, 0, 0]
        model_dict["p_RtL"] = [0, 0, 0]
        model_dict["p_CtL"] = [0, 0, 0]
        model_dict["name"] += " no_transient_vs"
        return type(self)(**model_dict)

    @staticmethod
    @validate_call
    def calc_k(kdash: PositiveFloat, c: Annotated[float, Gt(0), Le(1)]) -> float:
        """Translate between k & k'.

        As explained below equation 4 in paper: k' = k / (c * (c - 1))
        """
        return kdash * c * (1 - c)

    @property
    def kdash_(self) -> float:
        return self.k / (self.p_rce * (self.p_rce - 1))

    @validate_call
    def calc_R_leak_capacitor(
        self,
        duration: timedelta,
        SoC_final: Annotated[float, Ge(0), Le(1)],
        SoC_0: Annotated[float, Ge(0), Le(1)] = 1.0,
    ) -> float:
        # based on capacitor discharge: U(t) = U0 * e ^ (-t/RC)
        # Example: 50mAh; SoC from 100 % to 85 % over 30 days => ~1.8 MOhm
        U0 = self.calc_V_OC(SoC_0)
        Ut = self.calc_V_OC(SoC_final)
        return duration.total_seconds() * U0 / (self.q_As * math.log(U0 / Ut))

    @validate_call
    def calc_R_leak_battery(
        self,
        duration: timedelta,
        SoC_final: Annotated[float, Ge(0), Le(1)],
        SoC_0: Annotated[float, Ge(0), Le(1)] = 1.0,
    ) -> float:
        U0 = self.calc_V_OC(SoC_0)
        U1 = self.calc_V_OC(SoC_final)
        current_A = (SoC_0 - SoC_final) * self.q_As / duration.total_seconds()
        return (U0 + U1) / 2 / current_A

    def calc_V_OC(self, SoC: float) -> float:
        return (
            self.p_VOC[0] * math.pow(math.e, -self.p_VOC[1] * SoC)
            + self.p_VOC[2]
            + self.p_VOC[3] * SoC
            - self.p_VOC[4] * SoC**2
            + self.p_VOC[5] * SoC**3
        )

    @property
    def capacity_in_uF(self) -> float:
        return 1e6 * self.q_As / self.calc_V_OC(SoC=1.0)

    @property
    def charge_in_mAh(self) -> float:
        return 1e3 * self.q_As / 3600

    @property
    def V_init(self) -> float:
        return self.calc_V_OC(SoC=self.SoC_init)

    def calc_R_series(self, SoC: float) -> float:
        return (
            self.p_Rs[0] * math.pow(math.e, -self.p_Rs[1] * SoC)
            + self.p_Rs[2]
            + self.p_Rs[3] * SoC
            - self.p_Rs[4] * SoC**2
            + self.p_Rs[5] * SoC**3
        )

    def calc_R_transient_S(self, SoC: float) -> float:
        return self.p_RtS[0] * math.pow(math.e, -self.p_RtS[1] * SoC) + self.p_RtS[2]

    def calc_C_transient_S(self, SoC: float) -> float:
        return self.p_CtS[0] * math.pow(math.e, -self.p_CtS[1] * SoC) + self.p_CtS[2]

    def calc_R_transient_L(self, SoC: float) -> float:
        return self.p_RtL[0] * math.pow(math.e, -self.p_RtL[1] * SoC) + self.p_RtL[2]

    def calc_C_transient_L(self, SoC: float) -> float:
        return self.p_CtL[0] * math.pow(math.e, -self.p_CtL[1] * SoC) + self.p_CtL[2]

    def approximate_SoC(self, V_OC: float) -> float:
        SoC_next = SoC_now = 0.5
        step_size = 0.05
        go_up = True
        match = 5
        counter = 0

        while match > 0.001:
            SoC_now = SoC_next
            V_OC_now = self.calc_V_OC(SoC_now)
            match_new = abs(V_OC_now / V_OC - 1)
            if match_new > match:
                go_up = not go_up
                if go_up:
                    step_size /= 2
            SoC_next += step_size if go_up else -step_size
            match = match_new
            counter += 1
            if counter > 100:
                raise RuntimeError("Could not approximate SoC for given Voltage")

        return SoC_now


# constants & custom types
TIMESTEP_s_DEFAULT: float = 1.0 / config.SAMPLERATE_SPS
LuT_SIZE_LOG: int = 7
LuT_SIZE: int = 2**LuT_SIZE_LOG
u32 = Annotated[int, Field(ge=0, lt=2**32)]
lut_storage = Annotated[list[u32], Field(min_length=LuT_SIZE, max_length=LuT_SIZE)]


class StoragePRUConfig(ShpModel):
    """Map settings-list to internal state-vars struct StorageConfig.

    NOTE:
      - yaml is based on si-units like nA, mV, ms, uF
      - c-code and py-copy is using nA, uV, ns, nF, fW, raw
      - ordering is intentional and in sync with shepherd/commons.h
    """

    SoC_init_1_n30: u32
    """ ⤷ initial charge of storage """
    Constant_1_per_nA_n60: u32
    """ ⤷ Convert I_charge to delta-SoC with one multiplication."""
    Constant_1_per_uV_n60: u32
    """ ⤷ Leakage - Convert V_OC to delta-SoC with one multiplication.
    Combines prior constant and R_leak, to maximize resolution.
    """
    LuT_VOC_uV_n8: lut_storage
    """⤷ ranges from 3.9 uV to 16.7 V"""
    LuT_RSeries_kOhm_n32: lut_storage
    """⤷ ranges from 233n to 1 kOhm"""

    @classmethod
    @validate_call
    def from_vstorage(
        cls,
        data: VirtualStorageConfig | None,
        dt_s: PositiveFloat = TIMESTEP_s_DEFAULT,
        *,
        optimize_clamp: bool = True,
    ) -> Self:
        x_off = 0.5 if optimize_clamp else 0.0
        SoC_min = 1.0 / LuT_SIZE
        if data is None:
            data = VirtualStorageConfig.capacitor(C_uF=100, V_rated=10)
        V_OC_LuT = [data.calc_V_OC(SoC_min * (x + x_off)) for x in range(LuT_SIZE)]
        R_series_LuT = [data.calc_R_series(SoC_min * (x + x_off)) for x in range(LuT_SIZE)]
        Constant_1_per_A: float = dt_s / data.q_As
        Constant_1_per_V: float = Constant_1_per_A / data.R_leak_Ohm
        return cls(
            SoC_init_1_n30=round(2**30 * data.SoC_init),
            Constant_1_per_nA_n60=round((2**60 / 1e9) * Constant_1_per_A),
            Constant_1_per_uV_n60=round((2**60 / 1e6) * Constant_1_per_V),
            LuT_VOC_uV_n8=[round((2**8 * 1e6) * y) for y in V_OC_LuT],
            LuT_RSeries_kOhm_n32=[round((2**32 * 1e-3) * y) for y in R_series_LuT],
        )



from pathlib import Path

import pytest

from shepherd_core import fw_tools

from .conftest import files_elf


@pytest.fixture
def path_hex(tmp_path: Path) -> Path:
    path_elf = files_elf[0]
    path_hex = (tmp_path / (path_elf.stem + ".hex")).resolve()
    return fw_tools.elf_to_hex(path_elf, path_hex)


@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_elf_to_hex(path_elf: Path, tmp_path: Path) -> None:
    path_hex = (tmp_path / (path_elf.stem + ".hex")).resolve()
    path_gen = fw_tools.elf_to_hex(path_elf, path_hex)
    assert path_hex.exists()
    assert path_hex.as_posix() == path_gen.as_posix()


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_firmware_to_hex_w_elf(path_elf: Path) -> None:
    path_gen = fw_tools.firmware_to_hex(path_elf)
    assert path_gen.exists
    assert path_gen.suffix.lower() == ".hex"


@pytest.mark.elf
@pytest.mark.converter
def test_firmware_to_hex_w_hex(path_hex: Path) -> None:
    path_gen = fw_tools.firmware_to_hex(path_hex)
    assert path_gen.exists
    assert path_gen.suffix.lower() == ".hex"
    assert path_gen.as_posix() == path_hex.as_posix()


@pytest.mark.elf
@pytest.mark.converter
def test_firmware_to_hex_w_fail() -> None:
    path_some = Path(__file__).parent / "conftest.py"
    with pytest.raises(FileNotFoundError):
        _ = fw_tools.firmware_to_hex(path_some)


@pytest.mark.elf
@pytest.mark.converter
def test_hash() -> None:
    hash_a = fw_tools.file_to_hash(files_elf[0])
    hash_b = fw_tools.file_to_hash(files_elf[1])
    hash_c = fw_tools.file_to_hash(files_elf[1])
    assert hash_a != hash_b
    assert hash_b == hash_c


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_base64(path_elf: Path, tmp_path: Path) -> None:
    b64_a = fw_tools.file_to_base64(path_elf)
    path_b = (tmp_path / path_elf.name).resolve()
    fw_tools.base64_to_file(b64_a, path_b)
    b64_b = fw_tools.file_to_base64(path_b)
    assert b64_a == b64_b
    hash_a = fw_tools.file_to_hash(path_elf)
    hash_b = fw_tools.file_to_hash(path_b)
    assert hash_a == hash_b


# extract_firmware() is indirectly tested with Firmware-Class



"""Configuration related to Target Nodes (DuT)."""

from typing import Annotated

from pydantic import Field
from pydantic import model_validator
from typing_extensions import Self

from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.content.energy_environment import EnergyEnvironment
from shepherd_core.data_models.content.firmware import Firmware
from shepherd_core.data_models.content.virtual_source_config import VirtualSourceConfig
from shepherd_core.data_models.testbed.target import IdInt16
from shepherd_core.data_models.testbed.target import Target

from .observer_features import GpioActuation
from .observer_features import GpioTracing
from .observer_features import PowerTracing
from .observer_features import UartLogging

# defaults (pre-init complex types)
vsrc_neutral = VirtualSourceConfig(name="neutral")


class TargetConfig(ShpModel, title="Target Config"):
    """Configuration related to Target Nodes (DuT)."""

    target_IDs: Annotated[list[IdInt], Field(min_length=1, max_length=128)]
    custom_IDs: Annotated[list[IdInt16], Field(min_length=1, max_length=128)] | None = None
    """ ⤷ custom ID will replace 'const uint16_t SHEPHERD_NODE_ID' in firmware.

    if no custom ID is provided, the original ID of target is used
    """

    energy_env: EnergyEnvironment
    """ input for the virtual source """
    virtual_source: VirtualSourceConfig = vsrc_neutral
    target_delays: (
        Annotated[list[Annotated[int, Field(ge=0)]], Field(min_length=1, max_length=128)] | None
    ) = None
    """ ⤷ individual starting times

    - allows to use the same environment
    - not implemented ATM
    """

    firmware1: Firmware
    """ ⤷ omitted FW gets set to neutral deep-sleep"""
    firmware2: Firmware | None = None
    """ ⤷ omitted FW gets set to neutral deep-sleep"""

    power_tracing: PowerTracing | None = None
    gpio_tracing: GpioTracing | None = None
    gpio_actuation: GpioActuation | None = None
    uart_logging: UartLogging | None = None

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        if not self.energy_env.valid:
            msg = f"EnergyEnv '{self.energy_env.name}' for target must be valid"
            raise ValueError(msg)
        for _id in self.target_IDs:
            target = Target(id=_id)
            for mcu_num in [1, 2]:
                val_fw = getattr(self, f"firmware{mcu_num}")
                has_fw = val_fw is not None
                tgt_mcu = target[f"mcu{mcu_num}"]
                has_mcu = tgt_mcu is not None
                if not has_fw and has_mcu:
                    fw_def = Firmware(name=tgt_mcu.fw_name_default)
                    # ⤷ this will raise if default is faulty
                    if tgt_mcu.id != fw_def.mcu.id:
                        msg = (
                            f"Default-Firmware for MCU{mcu_num} of Target-ID '{target.id}' "
                            f"(={fw_def.mcu.name}) "
                            f"is incompatible (={tgt_mcu.name})"
                        )
                        raise ValueError(msg)
                if has_fw and has_mcu and val_fw.mcu.id != tgt_mcu.id:
                    msg = (
                        f"Firmware{mcu_num} for MCU of Target-ID '{target.id}' "
                        f"(={val_fw.mcu.name}) "
                        f"is incompatible (={tgt_mcu.name})"
                    )
                    raise ValueError(msg)

        c_ids = self.custom_IDs
        t_ids = self.target_IDs
        if c_ids is not None and (len(set(c_ids)) < len(set(t_ids))):
            msg = f"Provided custom IDs {c_ids} not enough to cover target range {t_ids}"
            raise ValueError(msg)
        # TODO: if custom ids present, firmware must be ELF
        if self.gpio_actuation is not None:
            raise NotImplementedError("Feature GpioActuation reserved for future use.")
        return self

    def get_custom_id(self, target_id: int) -> int | None:
        if self.custom_IDs is not None and target_id in self.target_IDs:
            return self.custom_IDs[self.target_IDs.index(target_id)]
        return None



import os
from pathlib import Path

import pytest

from shepherd_data import ivonne
from shepherd_data import mppt


@pytest.fixture
def file_ivonne() -> Path:
    path = Path(__file__).resolve().parent.parent.parent.parent / "shepherd_data/examples"
    os.chdir(path)
    return path / "./jogging_10m.iv"


@pytest.fixture
def file_isc_voc(file_ivonne: Path) -> Path:
    path = file_ivonne.parent / "jogging_10m_isc_voc.h5"
    if not path.exists():
        with ivonne.Reader(file_ivonne) as db:
            db.upsample_2_isc_voc(path, duration_s=1)
    return path


@pytest.fixture
def file_ivcurve(file_ivonne: Path) -> Path:
    path = file_ivonne.parent / "jogging_10m_ivcurve.h5"
    if not path.exists():
        with ivonne.Reader(file_ivonne) as db:
            db.convert_2_ivsurface(path, duration_s=1)
    return path


@pytest.fixture
def file_ivsample(file_ivonne: Path) -> Path:
    path = file_ivonne.parent / "jogging_10m_ivsample.h5"
    if not path.exists():
        with ivonne.Reader(file_ivonne) as db:
            tr_opt = mppt.OptimalTracker()
            db.convert_2_ivtrace(path, tracker=tr_opt, duration_s=1)
    return path


@pytest.fixture
def _file_cleanup(file_isc_voc: Path, file_ivcurve: Path, file_ivsample: Path) -> None:
    file_isc_voc.unlink(missing_ok=True)
    file_ivcurve.unlink(missing_ok=True)
    file_ivsample.unlink(missing_ok=True)



"""meta-data representation of a testbed-component (physical object)."""

from enum import Enum
from typing import Annotated
from typing import Any

from pydantic import Field
from pydantic import model_validator

from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.content import NameStr
from shepherd_core.data_models.base.content import SafeStr
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.testbed_client import tb_client


class ProgrammerProtocol(str, Enum):
    """Options regarding the programming-protocol."""

    SWD = swd = "SWD"
    SBW = sbw = "SBW"
    JTAG = jtag = "JTAG"
    UART = uart = "UART"


class MCU(ShpModel, title="Microcontroller of the Target Node"):
    """meta-data representation of a testbed-component (physical object)."""

    id: IdInt
    name: NameStr
    description: SafeStr
    comment: SafeStr | None = None

    platform: NameStr
    core: NameStr
    prog_protocol: ProgrammerProtocol
    prog_voltage: Annotated[float, Field(ge=1, le=5)] = 3
    prog_datarate: Annotated[int, Field(gt=0, le=1_000_000)] = 500_000

    fw_name_default: str
    # ⤷ can't be FW-Object (circular import)

    def __str__(self) -> str:
        return self.name

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        values, _ = tb_client.try_completing_model(cls.__name__, values)
        return values



"""UART - Decoder.

Features:

- 1 bit Start (LOW)
- 1 .. **8** .. 64 bit data-frame
- **1**, 1.5, 2 bit stop (HIGH) - don't care?
- **no**, even, odd parity bit (there are more, see sigrok-link) - TODO
- **LSB** / MSB first - detectable with dict-compare - TODO
- **no** inversion

Todo:
- detect bitOrder
- detect dataframe length
- detect parity

More Info:

https://en.wikipedia.org/wiki/Universal_asynchronous_receiver-transmitter
https://sigrok.org/wiki/Protocol_decoder:Uart

"""

from enum import Enum
from pathlib import Path

import numpy as np

from shepherd_core.logger import log


class Parity(str, Enum):
    """Options for parity-property of UART data-frame."""

    no = "no"
    even = "odd"
    odd = "even"


class BitOrder(str, Enum):
    """Options for bit-order-property of UART data-frame."""

    msb = msb_first = "msb"
    lsb = lsb_first = "lsb"


class Uart:
    """Specialized UART decoder."""

    def __init__(
        self,
        content: Path | np.ndarray,
        *,
        baud_rate: int | None = None,
        frame_length: int | None = 8,
        inversion: bool | None = None,
        parity: Parity | None = Parity.no,
        bit_order: BitOrder | None = BitOrder.lsb_first,
    ) -> None:
        """Provide a file with two columns: TS & Signal.

        Constraints:
        - timestamp -> seconds with fraction
        - signal -> can be analog or digital

        Note: class-parameters that are None (above) get auto-detected
          (some detectors still missing).
        """
        if isinstance(content, Path):
            self.events_sig: np.ndarray = np.loadtxt(content.as_posix(), delimiter=",", skiprows=1)
            # TODO: if float fails load as str -
            #  cast first col as np.datetime64 with ns-resolution, convert to delta
        else:
            self.events_sig = content

        # verify table
        if self.events_sig.shape[1] != 2:
            raise TypeError("Input file should have 2 rows -> (comma-separated) timestamp & value")
        if self.events_sig.shape[0] < 8:
            raise TypeError("Input file is too short (< state-changes)")
        # verify timestamps
        time_steps = self.events_sig[1:, 0] - self.events_sig[:-1, 0]
        if any(time_steps < 0):
            raise TypeError("Timestamps are not continuous")

        # prepare samples & process params (order is important)
        self._convert_analog2digital()
        self._filter_redundant_states()

        self.baud_rate: int = baud_rate if baud_rate is not None else self.detect_baud_rate()
        self.dur_tick: float = 1.0 / self.baud_rate

        self._add_duration()

        self.inversion: bool = inversion if inversion is not None else self.detect_inversion()
        self.half_stop: bool = self.detect_half_stop()  # not needed ATM

        # TODO: add detectors
        self.parity: Parity = parity if parity is not None else Parity.no
        self.bit_order: BitOrder = bit_order if bit_order is not None else BitOrder.lsb_first
        self.frame_length: int = frame_length if frame_length is not None else 8

        if not (0 < self.frame_length <= 64):
            raise ValueError("Dataframe length is out of bound (0, 64]")
        if self.parity != Parity.no:
            raise ValueError("Parities beside 'no' are not supported ATM")
        if self.bit_order != BitOrder.lsb_first:
            raise ValueError("Only bit-order LSB-first is supported ATM")

        if self.inversion:
            log.debug("inversion was detected / issued -> will invert signal")
            self._convert_analog2digital(invert=True)
        if self.detect_inversion():
            log.error("Signal still inverted?!? Check parameters and input")

        # results
        self.events_symbols: np.ndarray | None = None
        self.events_lines: np.ndarray | None = None
        self.text: str | None = None

    def _convert_analog2digital(self, *, invert: bool = False) -> None:
        """Divide dimension in two, divided by mean-value."""
        data = self.events_sig[:, 1]
        mean = np.mean(data)
        if invert:
            self.events_sig[:, 1] = data <= mean
        else:
            self.events_sig[:, 1] = data >= mean

    def _filter_redundant_states(self) -> None:
        """Sum of two sequential states is always 1 (True + False) if alternating."""
        data_0 = self.events_sig[:, 1]
        data_1 = np.concatenate([[not data_0[0]], data_0[:-1]])
        data_f = data_0 + data_1
        self.events_sig = self.events_sig[data_f == 1]

        if len(data_0) > len(self.events_sig):
            log.debug(
                "filtered out %d/%d events (redundant)",
                len(data_0) - len(self.events_sig),
                len(data_0),
            )

    def _add_duration(self) -> None:
        """Calculate third column -> duration of state in [baud-ticks]."""
        if self.events_sig.shape[1] > 2:
            log.warning("Tried to add state-duration, but it seems already present")
            return
        if not hasattr(self, "dur_tick"):
            raise ValueError("Make sure that baud-rate was calculated before running add_dur()")
        dur_steps = self.events_sig[1:, 0] - self.events_sig[:-1, 0]
        dur_steps = np.reshape(dur_steps, (dur_steps.size, 1))
        self.events_sig = np.append(self.events_sig[:-1, :], dur_steps / self.dur_tick, axis=1)

    def detect_inversion(self) -> bool:
        """Analyze bit-state during long pauses (unchanged states).

        - pause should be HIGH for non-inverted mode (default)
        - assumes maximum frame size of 64 bit + x for safety
        """
        events = self.events_sig[:1000, :]  # speedup for large datasets
        pauses = events[:, 2] > 80
        states_1 = events[:, 1]
        pause_states = states_1[pauses]
        mean_state = pause_states.mean()
        if 0.1 < mean_state < 0.9:
            raise ValueError("Inversion in pauses could not be detected")
        return mean_state < 0.5

    def detect_baud_rate(self) -> int:
        """Analyze the smallest step."""
        events = self.events_sig[:1000, :]  # speedup for large datasets
        dur_steps = events[1:, 0] - events[:-1, 0]
        def_step = np.percentile(dur_steps[dur_steps > 0], 10)
        mean_tick = dur_steps[
            (dur_steps >= 0.66 * def_step) & (dur_steps <= 1.33 * def_step)
        ].mean()
        return round(1 / mean_tick)

    def detect_half_stop(self) -> bool:
        """Look into the spacing between time-steps to determine use of half stop."""
        events = self.events_sig[:1000, :]  # speedup for large datasets
        return np.sum((events > 1.333 * self.dur_tick) & (events < 1.667 * self.dur_tick)) > 0

    def detect_dataframe_length(self) -> int:
        """Try to determine length of dataframe.

        Algo will look for longest pauses & accumulate steps until
        a state with uneven step-size is found.
        """
        raise NotImplementedError

    def get_symbols(self, *, force_redo: bool = False) -> np.ndarray:
        """Extract symbols from events.

        Ways to detect EOF:
        - long pause on HIGH
        - off_tick pause on high
        - bit_pos > max

        # TODO:
            - slowest FN -> speedup with numba or parallelization?
            - dset could be divided (long pauses) and threaded for speedup.
        """
        if force_redo:
            self.events_symbols = None
        if self.events_symbols is not None:
            return self.events_symbols

        pos_df: int | None = None
        symbol: int = 0
        t_start: float | None = None
        content: list = []

        for time, value, steps in self.events_sig:
            if steps > self.frame_length:
                if value:
                    # long pause on High -> reset frame
                    if pos_df is not None:
                        content.append([t_start, chr(symbol)])
                        t_start = None
                        symbol = 0
                    pos_df = None
                else:
                    log.debug("Error - Long pause - but SigLow (@%d)", time)
                continue
            if pos_df is None and value == 0:
                # Start of frame (first low after pause / EOF)
                pos_df = 0
                steps -= 1  # noqa: PLW2901
                t_start = time
            if pos_df is not None:
                if round(steps) >= 1 and value:
                    chunk = min(steps, self.frame_length - pos_df - 1)
                    lshift = min(pos_df, self.frame_length - 1)
                    symbol += (2 ** round(chunk) - 1) << lshift
                pos_df += round(steps)
                off_tick = abs(steps - round(steps)) > 0.1
                if pos_df >= self.frame_length or (off_tick and value):
                    # end of frame -> reset
                    if pos_df is not None:
                        content.append((t_start, chr(symbol)))
                        t_start = None
                        symbol = 0
                    pos_df = None
                    if off_tick and value == 0:
                        log.debug("Error - Off-sized step - but SigLow (@%d)", time)
        self.events_symbols = np.concatenate(content).reshape((len(content), 2))
        # TODO: numpy is converting timestamp to string -> must be added as tuple (ts, symbol)
        # symbol_events[:, 0] = symbol_events[:, 0].astype(float)  # noqa: ERA001
        # ⤷ does not work
        return self.events_symbols

    def get_lines(self, *, force_redo: bool = False) -> np.ndarray:
        r"""Timestamped symbols to line, cut at \r, \r\n or \n."""
        if force_redo:
            self.events_lines = None
        if self.events_lines is not None:
            return self.events_lines
        if self.events_symbols is None:
            self.get_symbols()

        content = []
        semi_stop = False
        t_start = None
        line = ""
        for time, symbol in self.events_symbols:
            if symbol == "\n" or semi_stop:
                if symbol == "\n":
                    line += symbol
                content.append((t_start, line))
                semi_stop = False
                t_start = None
                line = ""
                if not semi_stop:
                    continue
            if symbol == "\r":
                semi_stop = True

            if t_start is None:
                t_start = time
            line += symbol
        self.events_lines = np.concatenate(content).reshape((len(content), 2))
        return self.events_lines

    def get_text(self, *, force_redo: bool = False) -> str:
        """Remove timestamps and just return the whole string."""
        if force_redo:
            self.text = None
        if self.text is not None:
            return self.text
        if self.events_lines is None:
            self.get_lines()
        self.text = "".join(self.events_lines[:, 1])
        return self.text



"""meta-data representation of a testbed-component (physical object)."""

from datetime import datetime
from typing import Annotated
from typing import Any

from pydantic import Field
from pydantic import model_validator

from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.content import NameStr
from shepherd_core.data_models.base.content import SafeStr
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.testbed_client import tb_client

from .mcu import MCU

IdInt16 = Annotated[int, Field(ge=0, lt=2**16)]

MCUPort = Annotated[int, Field(ge=1, le=2)]


class Target(ShpModel, title="Target Node (DuT)"):
    """meta-data representation of a testbed-component (physical object)."""

    id: IdInt
    name: NameStr
    version: NameStr
    description: SafeStr

    comment: SafeStr | None = None

    active: bool = True
    created: datetime = Field(default_factory=datetime.now)

    testbed_id: IdInt16 | None = None
    """ ⤷ is derived from ID (targets are still selected by id!)"""
    mcu1: MCU | NameStr
    mcu2: MCU | NameStr | None = None

    # TODO: programming pins per mcu should be here (or better in Cape)

    def __str__(self) -> str:
        return self.name

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        values, _ = tb_client.try_completing_model(cls.__name__, values)

        # post correction
        for _mcu in ["mcu1", "mcu2"]:
            if isinstance(values.get(_mcu), str):
                values[_mcu] = MCU(name=values[_mcu])
                # ⤷ this will raise if default is faulty
            elif isinstance(values.get(_mcu), dict):
                values[_mcu] = MCU(**values[_mcu])
        if values.get("testbed_id") is None:
            values["testbed_id"] = values.get("id") % 2**16

        return values
